import duckdb
import numpy as np
import pandas as pd
import pyarrow as pa
from pytest import raises
from triad.utils.pyarrow import TRIAD_DEFAULT_TIMESTAMP

from fugue_duckdb._utils import encode_value_to_expr, to_duck_type, to_pa_type


def test_encode_value_to_expr():
    assert "1.1" == encode_value_to_expr(1.1)
    assert "1.1" == encode_value_to_expr(np.float64(1.1))
    assert "1" == encode_value_to_expr(1)
    assert "1" == encode_value_to_expr(np.int32(1))
    assert "FALSE" == encode_value_to_expr(False)
    assert "TRUE" == encode_value_to_expr(np.bool_(1))
    assert "E'abc'" == encode_value_to_expr("abc")
    assert "E'abc\\n;def'" == encode_value_to_expr("abc\n;def")
    assert "'\\xcaABC'::BLOB" == encode_value_to_expr(b"\xCAABC")
    assert "NULL" == encode_value_to_expr(None)
    assert "NULL" == encode_value_to_expr(float("nan"))
    assert "NULL" == encode_value_to_expr(pd.NA)
    assert "TIMESTAMP '2021-01-02 14:15:16'" == encode_value_to_expr(
        pd.to_datetime("2021-01-02 14:15:16")
    )
    assert "DATE '2021-01-02'" == encode_value_to_expr(
        pd.to_datetime("2021-01-02 14:15:16").date()
    )
    assert "[E'abc', 1, TIMESTAMP '2021-01-02 14:15:16']" == encode_value_to_expr(
        ["abc", 1, pd.to_datetime("2021-01-02 14:15:16")]
    )
    assert "{E'a\\n': E'abc', E'b': 1}" == encode_value_to_expr({"a\n": "abc", "b": 1})

    class D:
        pass

    raises(NotImplementedError, lambda: encode_value_to_expr(D()))


def test_type_conversion():
    con = duckdb.connect()

    def assert_(tp):
        dt = duckdb.arrow(pa.Table.from_pydict(dict(a=pa.nulls(2, tp))), con).types[0]
        assert to_pa_type(dt) == tp
        dt = to_duck_type(tp)
        assert to_pa_type(dt) == tp

    def assert_many(*tps):
        for tp in tps:
            assert_(tp)
            assert_(pa.list_(tp))

    assert_(pa.bool_())
    assert_many(pa.int8(), pa.int16(), pa.int32(), pa.int64())
    assert_many(pa.uint8(), pa.uint16(), pa.uint32(), pa.uint64())
    assert_many(pa.float32(), pa.float64())
    assert_(TRIAD_DEFAULT_TIMESTAMP)
    assert_(pa.decimal128(2, 3))
    assert_(pa.binary())
    assert_(pa.string())

    # nested
    assert_(pa.struct([pa.field("x", pa.int64())]))
    assert_(pa.struct([pa.field("x", pa.int64()), pa.field("yy", pa.string())]))
    assert_(pa.list_(pa.struct([pa.field("x", pa.int64())])))
    assert_(
        pa.struct([pa.field("x", pa.int64()), pa.field("yy", pa.list_(pa.int64()))])
    )
    assert_(
        pa.struct(
            [
                pa.field("x", pa.int64()),
                pa.field("yy", pa.struct([pa.field("x", pa.int64())])),
            ]
        )
    )
    assert_(
        pa.struct(
            [
                pa.field("x", pa.int64()),
                pa.field("yy", pa.list_(pa.struct([pa.field("x", pa.int64())]))),
            ]
        )
    )

    assert_(
        pa.map_(
            pa.struct(
                [
                    pa.field("x", pa.int64()),
                    pa.field("yy", pa.struct([pa.field("x", pa.int64())])),
                ]
            ),
            pa.struct(
                [
                    pa.field("yy", pa.list_(pa.struct([pa.field("x", pa.int64())]))),
                ]
            ),
        )
    )

    raises(ValueError, lambda: to_pa_type(""))
    raises(ValueError, lambda: to_pa_type("XX"))
    raises(ValueError, lambda: to_pa_type("XX(1,2)"))
    raises(ValueError, lambda: to_pa_type("XX<VARCHAR>"))



from typing import Type

from pytest import raises
from triad.collections.dict import ParamDict
from triad.utils.convert import get_full_type_path

from fugue import ExecutionEngine, NativeExecutionEngine, register_global_conf
from fugue.constants import FUGUE_CONF_SQL_IGNORE_CASE
from fugue.rpc.base import NativeRPCServer
from fugue_duckdb import DuckDBEngine


class _MockSQLEngine(DuckDBEngine):
    @property
    def execution_engine_constraint(self) -> Type[ExecutionEngine]:
        return _MockExecutionEngine


class _MockExecutionEngine(NativeExecutionEngine):
    def __init__(self, conf=None):
        super().__init__(conf=conf)
        self._stop = 0

    def stop_engine(self):
        self._stop += 1

    def create_default_sql_engine(self):
        return _MockSQLEngine(self)


class _MockRPC(NativeRPCServer):
    _start = 0
    _stop = 0

    def __init__(self, conf):
        super().__init__(conf)
        _MockRPC._start = 0
        _MockRPC._stop = 0

    def start_handler(self):
        _MockRPC._start += 1

    def stop_handler(self):
        _MockRPC._stop += 1


def test_sql_engine_init():
    engine = _MockExecutionEngine()
    assert isinstance(engine.sql_engine, _MockSQLEngine)

    with raises(TypeError):
        _MockSQLEngine(NativeExecutionEngine())


def test_start_stop():
    conf = {"fugue.rpc.server": get_full_type_path(_MockRPC)}
    engine = _MockExecutionEngine(conf=conf)
    engine.stop()
    assert 1 == engine._stop
    engine.stop()  # stop will be called only once
    assert 1 == engine._stop


def test_global_conf():
    register_global_conf({"ftest.a": 1})
    engine = _MockExecutionEngine()
    assert 1 == engine.conf.get_or_throw("ftest.a", int)
    engine = _MockExecutionEngine({"ftest.a": 2})
    assert 2 == engine.conf.get_or_throw("ftest.a", int)
    assert not engine.conf.get_or_throw(FUGUE_CONF_SQL_IGNORE_CASE, bool)

    # with duplicated value but it's the same as existing ones
    register_global_conf({"ftest.a": 1, "ftest.b": 2}, on_dup=ParamDict.THROW)
    engine = _MockExecutionEngine()
    assert 1 == engine.conf.get_or_throw("ftest.a", int)
    assert 2 == engine.conf.get_or_throw("ftest.b", int)

    # transactional, of one value has problem, the whole conf will not be added
    with raises(ValueError):
        register_global_conf({"ftest.a": 2, "ftest.c": 3}, on_dup=ParamDict.THROW)
    assert 1 == engine.conf.get_or_throw("ftest.a", int)
    assert 2 == engine.conf.get_or_throw("ftest.b", int)
    assert "ftest.c" not in engine.conf



import json
from datetime import datetime
from typing import Any

import numpy as np
import pandas as pd
from fugue.dataframe import (
    IterableDataFrame,
    PandasDataFrame,
    ArrayDataFrame,
    LocalDataFrameIterableDataFrame,
)
from fugue.dataframe.utils import _df_eq as df_eq
from fugue_test.dataframe_suite import DataFrameTests
from pytest import raises
from triad.collections.schema import Schema, SchemaError
from triad.exceptions import InvalidOperationError


class LocalDataFrameIterableDataFrameTests(DataFrameTests.Tests):
    def df(self, data: Any = None, schema: Any = None) -> IterableDataFrame:
        def get_dfs():
            if isinstance(data, list):
                for row in data:
                    yield ArrayDataFrame([], schema)  # noise
                    yield ArrayDataFrame([row], schema)
                if schema is None:
                    yield ArrayDataFrame([], schema)  # noise
            elif data is not None:
                yield ArrayDataFrame(data, schema)

        return LocalDataFrameIterableDataFrame(get_dfs(), schema)


def test_init():
    raises(Exception, lambda: LocalDataFrameIterableDataFrame(1))

    def get_dfs(seq):
        for x in seq:
            if x == "e":
                yield IterableDataFrame([], "a:int,b:int")
            if x == "v":
                yield IterableDataFrame([[1, 10]], "a:int,b:int")
            if x == "o":  # bad schema but empty dataframe doesn't matter
                yield ArrayDataFrame([], "a:int,b:str")

    with raises(Exception):
        LocalDataFrameIterableDataFrame(get_dfs(""))

    with raises(Exception):  # schema mismatch
        LocalDataFrameIterableDataFrame(get_dfs("v"), "a:int,b:str")

    df = LocalDataFrameIterableDataFrame(get_dfs(""), "a:str")
    assert df.empty
    assert df.schema == "a:str"

    df = LocalDataFrameIterableDataFrame(get_dfs("e"))
    assert df.empty
    assert df.schema == "a:int,b:int"

    df = LocalDataFrameIterableDataFrame(get_dfs("eeeo"))
    assert df.empty
    assert df.schema == "a:int,b:str"

    df = LocalDataFrameIterableDataFrame(get_dfs("eveo"))
    assert not df.empty
    assert df.schema == "a:int,b:int"
    assert [[1, 10]] == df.as_array()

    df = LocalDataFrameIterableDataFrame(get_dfs("oveo"), "a:int,b:int")
    assert not df.empty
    assert df.schema == "a:int,b:int"
    assert [[1, 10]] == df.as_array()



from datetime import datetime
from typing import Any

import duckdb
import pandas as pd

import fugue.api as fa
from fugue import ArrowDataFrame
from fugue_duckdb.dataframe import DuckDataFrame
from fugue_test.dataframe_suite import DataFrameTests


class DuckDataFrameTests(DataFrameTests.Tests):
    @classmethod
    def setUpClass(cls):
        cls._con = duckdb.connect()

    def df(self, data: Any = None, schema: Any = None) -> DuckDataFrame:
        df = ArrowDataFrame(data, schema)
        return DuckDataFrame(duckdb.arrow(df.native, self._con))

    def test_as_array_special_values(self):
        for func in [
            lambda df, *args, **kwargs: df.as_array(*args, **kwargs, type_safe=True),
            lambda df, *args, **kwargs: list(
                df.as_array_iterable(*args, **kwargs, type_safe=True)
            ),
        ]:
            df = self.df([[pd.Timestamp("2020-01-01"), 1]], "a:datetime,b:int")
            data = func(df)
            assert [[datetime(2020, 1, 1), 1]] == data
            assert isinstance(data[0][0], datetime)
            assert isinstance(data[0][1], int)

            df = self.df([[pd.NaT, 1]], "a:datetime,b:int")
            assert [[None, 1]] == func(df)

            df = self.df([[float("nan"), 1]], "a:double,b:int")
            assert [[None, 1]] == func(df)

            # DuckDB disallows nan and inf
            # see https://github.com/duckdb/duckdb/pull/541

            # df = self.df([[float("inf"), 1]], "a:double,b:int")
            # assert [[float("inf"), 1]] == func(df)

    def test_as_pandas_duck(self):
        df = self.df([[2.1, 1]], "a:double,b:int")
        assert df.as_pandas().values.tolist() == [[2.1, 1]]
        df = self.df([[2.1, [1]]], "a:double,b:[int]")
        assert df.as_pandas().values.tolist() == [[2.1, [1]]]
        df = self.df([[2.1, ["a"]]], "a:double,b:[str]")
        assert df.as_pandas().values.tolist() == [[2.1, ["a"]]]
        df = self.df([[2.1, {"a": 1}]], "a:double,b:{a:int}")
        assert df.as_pandas().values.tolist() == [[2.1, {"a": 1}]]

    def test_init(self):
        df = self.df([], "a:int,b:str")
        assert df.schema == "a:int,b:str"
        assert df.empty
        assert isinstance(df.native, duckdb.DuckDBPyRelation)
        assert df.is_bounded
        assert df.is_local

    def test_duck_as_local(self):
        df = self.df([[2.1, 1]], "a:double,b:int")
        assert isinstance(df.as_local(), ArrowDataFrame)


class NativeDuckDataFrameTests(DataFrameTests.NativeTests):
    @classmethod
    def setUpClass(cls):
        cls._con = duckdb.connect()

    def df(self, data: Any = None, schema: Any = None) -> DuckDataFrame:
        df = ArrowDataFrame(data, schema)
        return DuckDataFrame(duckdb.arrow(df.native, self._con)).native

    def to_native_df(self, pdf: pd.DataFrame) -> Any:
        return duckdb.from_df(pdf)

    def test_num_partitions(self):
        assert fa.get_num_partitions(self.df([[0, 1]], "a:int,b:int")) == 1



import logging
import os
from typing import Any, Callable, Dict, List, Optional, Type, Union

import pandas as pd
from qpd_pandas import run_sql_on_pandas
from qpd_pandas.engine import PandasUtils
from triad import Schema
from triad.collections.dict import IndexedOrderedDict
from triad.collections.fs import FileSystem
from triad.utils.assertion import assert_or_throw

from fugue._utils.io import load_df, save_df
from fugue.collections.partition import (
    PartitionCursor,
    PartitionSpec,
    parse_presort_exp,
)
from fugue.collections.sql import StructuredRawSQL
from fugue.dataframe import (
    AnyDataFrame,
    DataFrame,
    DataFrames,
    LocalBoundedDataFrame,
    LocalDataFrame,
    PandasDataFrame,
    fugue_annotated_param,
)
from fugue.dataframe.dataframe import as_fugue_df
from fugue.dataframe.utils import get_join_schemas

from .execution_engine import (
    ExecutionEngine,
    ExecutionEngineParam,
    MapEngine,
    SQLEngine,
)


class QPDPandasEngine(SQLEngine):
    """QPD execution implementation.

    :param execution_engine: the execution engine this sql engine will run on
    """

    @property
    def dialect(self) -> Optional[str]:
        return "spark"

    def to_df(self, df: AnyDataFrame, schema: Any = None) -> DataFrame:
        return _to_native_execution_engine_df(df, schema)

    @property
    def is_distributed(self) -> bool:
        return False

    def select(self, dfs: DataFrames, statement: StructuredRawSQL) -> DataFrame:
        _dfs, _sql = self.encode(dfs, statement)
        _dd = {k: self.to_df(v).as_pandas() for k, v in _dfs.items()}  # type: ignore

        df = run_sql_on_pandas(_sql, _dd, ignore_case=True)
        return self.to_df(df)


class PandasMapEngine(MapEngine):
    @property
    def execution_engine_constraint(self) -> Type[ExecutionEngine]:
        return NativeExecutionEngine

    def to_df(self, df: AnyDataFrame, schema: Any = None) -> DataFrame:
        return _to_native_execution_engine_df(df, schema)

    @property
    def is_distributed(self) -> bool:
        return False

    def map_dataframe(
        self,
        df: DataFrame,
        map_func: Callable[[PartitionCursor, LocalDataFrame], LocalDataFrame],
        output_schema: Any,
        partition_spec: PartitionSpec,
        on_init: Optional[Callable[[int, DataFrame], Any]] = None,
        map_func_format_hint: Optional[str] = None,
    ) -> DataFrame:
        # if partition_spec.num_partitions != "0":
        #     self.log.warning(
        #         "%s doesn't respect num_partitions %s",
        #         self,
        #         partition_spec.num_partitions,
        #     )
        is_coarse = partition_spec.algo == "coarse"
        presort = partition_spec.get_sorts(df.schema, with_partition_keys=is_coarse)
        presort_keys = list(presort.keys())
        presort_asc = list(presort.values())
        output_schema = Schema(output_schema)
        cursor = partition_spec.get_cursor(df.schema, 0)
        if on_init is not None:
            on_init(0, df)
        if (
            len(partition_spec.partition_by) == 0 or partition_spec.algo == "coarse"
        ):  # no partition
            if len(partition_spec.presort) > 0:
                pdf = (
                    df.as_pandas()
                    .sort_values(presort_keys, ascending=presort_asc)
                    .reset_index(drop=True)
                )
                input_df = PandasDataFrame(pdf, df.schema, pandas_df_wrapper=True)
                cursor.set(lambda: input_df.peek_array(), cursor.partition_no + 1, 0)
                output_df = map_func(cursor, input_df)
            else:
                df = df.as_local()
                cursor.set(lambda: df.peek_array(), 0, 0)
                output_df = map_func(cursor, df)
            if (
                isinstance(output_df, PandasDataFrame)
                and output_df.schema != output_schema
            ):
                output_df = PandasDataFrame(output_df.native, output_schema)
            assert_or_throw(
                output_df.schema == output_schema,
                lambda: f"map output {output_df.schema} "
                f"mismatches given {output_schema}",
            )
            return self.to_df(output_df)  # type: ignore

        def _map(pdf: pd.DataFrame) -> pd.DataFrame:
            if len(partition_spec.presort) > 0:
                pdf = pdf.sort_values(presort_keys, ascending=presort_asc).reset_index(
                    drop=True
                )
            input_df = PandasDataFrame(pdf, df.schema, pandas_df_wrapper=True)
            cursor.set(lambda: input_df.peek_array(), cursor.partition_no + 1, 0)
            output_df = map_func(cursor, input_df)
            return output_df.as_pandas()

        result = self.execution_engine.pl_utils.safe_groupby_apply(  # type: ignore
            df.as_pandas(), partition_spec.partition_by, _map
        )
        return PandasDataFrame(result, output_schema)


class NativeExecutionEngine(ExecutionEngine):
    """The execution engine based on native python and pandas. This execution engine
    is mainly for prototyping and unit tests.

    Please read |ExecutionEngineTutorial| to understand this important Fugue concept

    :param conf: |ParamsLikeObject|, read |FugueConfig| to learn Fugue specific options
    """

    def __init__(self, conf: Any = None):
        super().__init__(conf)
        self._fs = FileSystem()
        self._log = logging.getLogger()

    def __repr__(self) -> str:
        return "NativeExecutionEngine"

    @property
    def log(self) -> logging.Logger:
        return self._log

    @property
    def fs(self) -> FileSystem:
        return self._fs

    @property
    def is_distributed(self) -> bool:
        return False

    def create_default_sql_engine(self) -> SQLEngine:
        return QPDPandasEngine(self)

    def create_default_map_engine(self) -> MapEngine:
        return PandasMapEngine(self)

    def get_current_parallelism(self) -> int:
        return 1

    @property
    def pl_utils(self) -> PandasUtils:
        """Pandas-like dataframe utils"""
        return PandasUtils()

    def to_df(self, df: AnyDataFrame, schema: Any = None) -> LocalBoundedDataFrame:
        return _to_native_execution_engine_df(df, schema)  # type: ignore

    def repartition(
        self, df: DataFrame, partition_spec: PartitionSpec
    ) -> DataFrame:  # pragma: no cover
        # self.log.warning("%s doesn't respect repartition", self)
        return df

    def broadcast(self, df: DataFrame) -> DataFrame:
        return self.to_df(df)

    def persist(
        self,
        df: DataFrame,
        lazy: bool = False,
        **kwargs: Any,
    ) -> DataFrame:
        return self.to_df(df)

    def join(
        self,
        df1: DataFrame,
        df2: DataFrame,
        how: str,
        on: Optional[List[str]] = None,
    ) -> DataFrame:
        key_schema, output_schema = get_join_schemas(df1, df2, how=how, on=on)
        d = self.pl_utils.join(
            df1.as_pandas(), df2.as_pandas(), join_type=how, on=key_schema.names
        )
        return PandasDataFrame(d.reset_index(drop=True), output_schema)

    def union(
        self,
        df1: DataFrame,
        df2: DataFrame,
        distinct: bool = True,
    ) -> DataFrame:
        assert_or_throw(
            df1.schema == df2.schema,
            lambda: ValueError(f"{df1.schema} != {df2.schema}"),
        )
        d = self.pl_utils.union(df1.as_pandas(), df2.as_pandas(), unique=distinct)
        return PandasDataFrame(d.reset_index(drop=True), df1.schema)

    def subtract(
        self,
        df1: DataFrame,
        df2: DataFrame,
        distinct: bool = True,
    ) -> DataFrame:
        assert_or_throw(
            distinct, NotImplementedError("EXCEPT ALL for NativeExecutionEngine")
        )
        assert_or_throw(
            df1.schema == df2.schema,
            lambda: ValueError(f"{df1.schema} != {df2.schema}"),
        )
        d = self.pl_utils.except_df(df1.as_pandas(), df2.as_pandas(), unique=distinct)
        return PandasDataFrame(d.reset_index(drop=True), df1.schema)

    def intersect(
        self,
        df1: DataFrame,
        df2: DataFrame,
        distinct: bool = True,
    ) -> DataFrame:
        assert_or_throw(
            distinct, NotImplementedError("INTERSECT ALL for NativeExecutionEngine")
        )
        assert_or_throw(
            df1.schema == df2.schema,
            lambda: ValueError(f"{df1.schema} != {df2.schema}"),
        )
        d = self.pl_utils.intersect(df1.as_pandas(), df2.as_pandas(), unique=distinct)
        return PandasDataFrame(d.reset_index(drop=True), df1.schema)

    def distinct(
        self,
        df: DataFrame,
    ) -> DataFrame:
        d = self.pl_utils.drop_duplicates(df.as_pandas())
        return PandasDataFrame(d.reset_index(drop=True), df.schema)

    def dropna(
        self,
        df: DataFrame,
        how: str = "any",
        thresh: Optional[int] = None,
        subset: List[str] = None,
    ) -> DataFrame:
        kwargs: Dict[str, Any] = dict(axis=0, subset=subset, inplace=False)
        if thresh is None:
            kwargs["how"] = how
        else:
            kwargs["thresh"] = thresh
        d = df.as_pandas().dropna(**kwargs)
        return PandasDataFrame(d.reset_index(drop=True), df.schema)

    def fillna(
        self,
        df: DataFrame,
        value: Any,
        subset: List[str] = None,
    ) -> DataFrame:
        assert_or_throw(
            (not isinstance(value, list)) and (value is not None),
            ValueError("fillna value can not None or a list"),
        )
        if isinstance(value, dict):
            assert_or_throw(
                (None not in value.values()) and (any(value.values())),
                ValueError(
                    "fillna dict can not contain None and needs at least one value"
                ),
            )
            mapping = value
        else:
            # If subset is none, apply to all columns
            subset = subset or df.columns
            mapping = {col: value for col in subset}
        d = df.as_pandas().fillna(mapping, inplace=False)
        return PandasDataFrame(d.reset_index(drop=True), df.schema)

    def sample(
        self,
        df: DataFrame,
        n: Optional[int] = None,
        frac: Optional[float] = None,
        replace: bool = False,
        seed: Optional[int] = None,
    ) -> DataFrame:
        assert_or_throw(
            (n is None and frac is not None) or (n is not None and frac is None),
            ValueError("one and only one of n and frac should be set"),
        )
        d = df.as_pandas().sample(n=n, frac=frac, replace=replace, random_state=seed)
        return PandasDataFrame(d.reset_index(drop=True), df.schema)

    def take(
        self,
        df: DataFrame,
        n: int,
        presort: str,
        na_position: str = "last",
        partition_spec: Optional[PartitionSpec] = None,
    ) -> DataFrame:
        partition_spec = partition_spec or PartitionSpec()
        assert_or_throw(
            isinstance(n, int),
            ValueError("n needs to be an integer"),
        )
        d = df.as_pandas()

        # Use presort over partition_spec.presort if possible
        if presort:
            presort = parse_presort_exp(presort)
        _presort: IndexedOrderedDict = presort or partition_spec.presort

        if len(_presort.keys()) > 0:
            d = d.sort_values(
                list(_presort.keys()),
                ascending=list(_presort.values()),
                na_position=na_position,
            )

        if len(partition_spec.partition_by) == 0:
            d = d.head(n)
        else:
            d = d.groupby(by=partition_spec.partition_by, dropna=False).head(n)

        return PandasDataFrame(
            d.reset_index(drop=True), df.schema, pandas_df_wrapper=True
        )

    def load_df(
        self,
        path: Union[str, List[str]],
        format_hint: Any = None,
        columns: Any = None,
        **kwargs: Any,
    ) -> LocalBoundedDataFrame:
        return self.to_df(
            load_df(
                path, format_hint=format_hint, columns=columns, fs=self.fs, **kwargs
            )
        )

    def save_df(
        self,
        df: DataFrame,
        path: str,
        format_hint: Any = None,
        mode: str = "overwrite",
        partition_spec: Optional[PartitionSpec] = None,
        force_single: bool = False,
        **kwargs: Any,
    ) -> None:
        partition_spec = partition_spec or PartitionSpec()
        if not force_single and not partition_spec.empty:
            kwargs["partition_cols"] = partition_spec.partition_by
        self.fs.makedirs(os.path.dirname(path), recreate=True)
        df = self.to_df(df)
        save_df(df, path, format_hint=format_hint, mode=mode, fs=self.fs, **kwargs)


@fugue_annotated_param(NativeExecutionEngine)
class _NativeExecutionEngineParam(ExecutionEngineParam):
    pass


def _to_native_execution_engine_df(df: AnyDataFrame, schema: Any = None) -> DataFrame:
    fdf = as_fugue_df(df) if schema is None else as_fugue_df(df, schema=schema)
    return fdf.as_local_bounded()



import pandas as pd
from fugue import register_execution_engine
from fugue import FugueSQLWorkflow
from pyspark.sql import SparkSession

from fugue_spark import SparkExecutionEngine


def test_sql(spark_session):
    register_execution_engine(
        "_spark",
        lambda conf, **kwargs: SparkExecutionEngine(
            conf=conf, spark_session=spark_session
        ),
    )
    df = spark_session.createDataFrame(pd.DataFrame([[0], [1]], columns=["a"]))
    dag = FugueSQLWorkflow()
    dag(
        """
    SELECT * FROM df WHERE a>0
    PRINT
    """,
        df=df,
    )
    dag.run("_spark")



import os

import numpy as np
import pandas as pd
import pyarrow as pa
from pytest import raises
from triad import FileSystem, Schema
from triad.collections.schema import SchemaError
from triad.exceptions import InvalidOperationError, NoneArgumentError

from fugue import ArrayDataFrame, IterableDataFrame, PandasDataFrame
from fugue.dataframe.utils import _df_eq as df_eq
from fugue.dataframe.utils import (
    _schema_eq,
    deserialize_df,
    get_column_names,
    get_join_schemas,
    normalize_dataframe_column_names,
    pickle_df,
    rename,
    serialize_df,
    unpickle_df,
)


def test_schema_eq():
    assert not _schema_eq(Schema("a:int"), Schema("a:int8"))
    assert not _schema_eq(Schema("a:int"), Schema("b:int"))
    assert not _schema_eq(Schema("a:int,b:int"), Schema("a:int"))

    f1 = pa.field("a", pa.list_(pa.field("x", pa.string())))
    f2 = pa.field("a", pa.list_(pa.field("y", pa.string())))
    s1 = Schema([f1, pa.field("b", pa.int64())])
    s2 = Schema([f2, pa.field("b", pa.int64())])
    assert _schema_eq(s1, s2)

    # nested
    s1 = Schema([pa.field("a", pa.list_(f1)), pa.field("b", pa.int64())])
    s2 = Schema([pa.field("a", pa.list_(f2)), pa.field("b", pa.int64())])
    assert _schema_eq(s1, s2)


def test_df_eq():
    df1 = ArrayDataFrame([[0, 100.0, "a"]], "a:int,b:double,c:str")
    df2 = ArrayDataFrame([[0, 100.001, "a"]], "a:int,b:double,c:str")
    assert df_eq(df1, df1)
    assert df_eq(df1, df2, digits=4)
    # precision
    assert not df_eq(df1, df2, digits=6)
    # no content
    assert df_eq(df1, df2, digits=6, check_content=False)
    raises(AssertionError, lambda: df_eq(df1, df2, throw=True))

    df1 = ArrayDataFrame([[100.0, "a"]], "a:double,b:str")
    assert df_eq(df1, df1.as_pandas(), df1.schema)

    df1 = ArrayDataFrame([[None, "a"]], "a:double,b:str")
    assert df_eq(df1, df1)

    df1 = ArrayDataFrame([[None, "a"]], "a:double,b:str")
    df2 = ArrayDataFrame([[np.nan, "a"]], "a:double,b:str")
    assert df_eq(df1, df2)

    df1 = ArrayDataFrame([[100.0, None]], "a:double,b:str")
    df2 = ArrayDataFrame([[100.0, None]], "a:double,b:str")
    assert df_eq(df1, df2)

    df1 = ArrayDataFrame([[0], [1]], "a:int")
    df2 = ArrayDataFrame([[1], [0]], "a:int")
    assert df_eq(df1, df2)
    assert not df_eq(df1, df2, check_order=True)


def test_get_join_schemas():
    a = ArrayDataFrame([], "a:int,b:int")
    b = ArrayDataFrame([], "c:int")
    c = ArrayDataFrame([], "d:str,a:int")
    i, u = get_join_schemas(a, b, how="cross", on=[])
    assert i == ""
    assert u == "a:int,b:int,c:int"
    raises(NoneArgumentError, lambda: get_join_schemas(a, b, how=None, on=[]))
    raises(ValueError, lambda: get_join_schemas(a, b, how="x", on=[]))
    raises(SchemaError, lambda: get_join_schemas(a, b, how="CROSS", on=["a"]))
    raises(SchemaError, lambda: get_join_schemas(a, c, how="CROSS", on=["a"]))
    raises(SchemaError, lambda: get_join_schemas(a, c, how="CROSS", on=[]))
    raises(SchemaError, lambda: get_join_schemas(a, b, how="inner", on=["a"]))
    raises(ValueError, lambda: get_join_schemas(a, c, how="outer", on=["a"]))
    i, u = get_join_schemas(a, c, how="inner", on=["a"])
    assert i == "a:int"
    assert u == "a:int,b:int,d:str"
    i, u = get_join_schemas(a, c, how="inner", on=[])  # infer
    assert i == "a:int"
    assert u == "a:int,b:int,d:str"
    a = ArrayDataFrame([], "a:int,b:int,c:int")
    b = ArrayDataFrame([], "c:int,b:int,x:int")
    raises(SchemaError, lambda: get_join_schemas(a, b, how="inner", on=["a"]))
    i, u = get_join_schemas(a, b, how="inner", on=["c", "b"])
    assert i == "b:int,c:int"
    assert u == "a:int,b:int,c:int,x:int"
    for how in ["SEMI", "LEFT_Semi", "Anti", "left_Anti"]:
        i, u = get_join_schemas(c, a, how=how, on=["a"])
        assert i == "a:int"
        assert u == "d:str,a:int"


def test_pickle_df():
    def assert_eq(df, df_expected=None, raw=False):
        if df_expected is None:
            df_expected = df
        df_actual = unpickle_df(pickle_df(df))
        if raw:
            assert df_expected.native == df_actual.native
        else:
            df_eq(df_expected, df_actual, throw=True)

    assert_eq(ArrayDataFrame([], "a:int,b:int"))
    assert_eq(ArrayDataFrame([[None, None]], "a:int,b:int"))
    assert_eq(ArrayDataFrame([[None, "abc"]], "a:int,b:str"))
    assert_eq(
        ArrayDataFrame([[None, [1, 2], dict(x=1)]], "a:int,b:[int],c:{x:int}"), raw=True
    )
    assert_eq(
        IterableDataFrame([[None, [1, 2], dict(x=1)]], "a:int,b:[int],c:{x:int}"),
        ArrayDataFrame([[None, [1, 2], dict(x=1)]], "a:int,b:[int],c:{x:int}"),
        raw=True,
    )
    assert_eq(PandasDataFrame([[None, None]], "a:int,b:int"))
    assert_eq(PandasDataFrame([[None, "abc"]], "a:int,b:str"))


def test_serialize_df(tmpdir):
    def assert_eq(df, df_expected=None, raw=False):
        if df_expected is None:
            df_expected = df
        df_actual = deserialize_df(serialize_df(df))
        if raw:
            assert df_expected.native == df_actual.native
        else:
            df_eq(df_expected, df_actual, throw=True)

    fs = FileSystem()
    assert deserialize_df(serialize_df(None)) is None
    assert_eq(ArrayDataFrame([], "a:int,b:int"))
    assert_eq(ArrayDataFrame([[None, None]], "a:int,b:int"))
    assert_eq(ArrayDataFrame([[None, "abc"]], "a:int,b:str"))
    assert_eq(
        ArrayDataFrame([[None, [1, 2], dict(x=1)]], "a:int,b:[int],c:{x:int}"), raw=True
    )
    assert_eq(
        IterableDataFrame([[None, [1, 2], dict(x=1)]], "a:int,b:[int],c:{x:int}"),
        ArrayDataFrame([[None, [1, 2], dict(x=1)]], "a:int,b:[int],c:{x:int}"),
        raw=True,
    )
    assert_eq(PandasDataFrame([[None, None]], "a:int,b:int"))
    assert_eq(PandasDataFrame([[None, "abc"]], "a:int,b:str"))

    raises(
        InvalidOperationError,
        lambda: serialize_df(ArrayDataFrame([], "a:int,b:int"), 0),
    )

    path = os.path.join(tmpdir, "1.pkl")

    df = ArrayDataFrame([[None, None]], "a:int,b:int")
    s = serialize_df(df, 0, path, fs)
    df_eq(df, deserialize_df(s, fs), throw=True)
    df_eq(df, deserialize_df(s), throw=True)

    s = serialize_df(df, 0, path)
    df_eq(df, deserialize_df(s), throw=True)

    raises(ValueError, lambda: deserialize_df('{"x":1}'))


def _test_get_column_names():
    df = pd.DataFrame([[0, 1, 2]])
    assert get_column_names(df) == [0, 1, 2]

    adf = pa.Table.from_pandas(df)
    assert get_column_names(adf) == ["0", "1", "2"]

    pdf = PandasDataFrame(pd.DataFrame([[0, 1]], columns=["a", "b"]))
    assert get_column_names(pdf) == ["a", "b"]


def _test_rename():
    assert rename("dummy", {}) == "dummy"
    pdf = pd.DataFrame([[0, 1, 2]], columns=["a", "b", "c"])
    df = rename(pdf, {})
    assert get_column_names(df) == ["a", "b", "c"]
    df = rename(pdf, {"b": "bb"})
    assert get_column_names(df) == ["a", "bb", "c"]

    adf = pa.Table.from_pandas(pdf)
    adf = rename(adf, {})
    assert get_column_names(adf) == ["a", "b", "c"]
    adf = rename(adf, {"b": "bb"})
    assert get_column_names(adf) == ["a", "bb", "c"]

    fdf = PandasDataFrame(pdf)
    fdf = rename(fdf, {})
    assert get_column_names(fdf) == ["a", "b", "c"]
    fdf = rename(fdf, {"b": "bb"})
    assert get_column_names(fdf) == ["a", "bb", "c"]


def test_normalize_dataframe_column_names():
    df = pd.DataFrame([[0, 1, 2]], columns=["a", "b", "c"])
    df, names = normalize_dataframe_column_names(df)
    assert get_column_names(df) == ["a", "b", "c"]
    assert names == {}

    df = pd.DataFrame([[0, 1, 2]])
    df, names = normalize_dataframe_column_names(df)
    assert get_column_names(df) == ["_0", "_1", "_2"]
    assert names == {"_0": 0, "_1": 1, "_2": 2}

    df = pd.DataFrame([[0, 1, 2, 3]], columns=["1", "2", "_2", "大"])
    df, names = normalize_dataframe_column_names(df)
    assert get_column_names(df) == ["_1", "_2_1", "_2", "_1_1"]
    assert names == {"_1": "1", "_2_1": "2", "_1_1": "大"}



from fugue.dataframe.pandas_dataframe import PandasDataFrame
from fugue_spark._utils.convert import to_schema, to_spark_schema
from fugue_spark._utils.partition import (
    even_repartition,
    hash_repartition,
    rand_repartition,
)
from pyspark.sql import SparkSession


def test_hash_repartition(spark_session):
    df = _df([[0, 1], [0, 2], [0, 3], [0, 4], [1, 1], [1, 2], [1, 3]], "a:int,b:int")
    res = hash_repartition(spark_session, df, 0, []).collect()
    assert 7 == len(res)
    res = hash_repartition(spark_session, df, 1, []).rdd.mapPartitions(_pc).collect()
    assert 7 == len(res)
    assert 7 == len([x for x in res if x[2] == 7])
    res = hash_repartition(spark_session, df, 1, ["a"]).rdd.mapPartitions(_pc).collect()
    assert 7 == len(res)
    assert 7 == len([x for x in res if x[2] == 7])
    res = hash_repartition(spark_session, df, 0, ["a"]).rdd.mapPartitions(_pc).collect()
    assert 7 == len(res)
    res = (
        hash_repartition(spark_session, df, 10, ["a"]).rdd.mapPartitions(_pc).collect()
    )
    assert 7 == len(res)
    assert 4 == len([x for x in res if x[2] == 4])
    assert 3 == len([x for x in res if x[2] == 3])


def test_rand_repartition(spark_session):
    df = _df([[0, 1], [0, 2], [0, 3], [0, 4], [1, 1], [1, 2], [1, 3]], "a:int,b:int")
    for i in [0, 1, 2, 3]:
        for p in [[], ["a"]]:
            res = (
                rand_repartition(spark_session, df, i, p)
                .rdd.mapPartitions(_pc)
                .collect()
            )
            assert 7 == len(res)


def test_even_repartition_no_cols(spark_session):
    df = _df([[0, 1], [0, 2], [0, 3], [0, 4], [1, 1], [1, 2], [1, 3]], "a:int,b:int")
    res = even_repartition(spark_session, df, 0, []).collect()
    assert 7 == len(res)
    res = even_repartition(spark_session, df, 1, []).rdd.mapPartitions(_pc).collect()
    assert 7 == len(res)
    assert 7 == len([x for x in res if x[2] == 7])
    res = even_repartition(spark_session, df, 6, []).rdd.mapPartitions(_pc).collect()
    assert 7 == len(res)
    assert 5 == len([x for x in res if x[2] == 1])
    assert 2 == len([x for x in res if x[2] == 2])
    res = even_repartition(spark_session, df, 7, []).rdd.mapPartitions(_pc).collect()
    assert 7 == len(res)
    assert all(x[2] == 1 for x in res)
    res = even_repartition(spark_session, df, 8, []).rdd.mapPartitions(_pc).collect()
    assert 7 == len(res)
    assert all(x[2] == 1 for x in res)


def test_even_repartition_with_cols(spark_session):
    df = _df([[0, 1], [0, 2], [0, 3], [0, 4], [1, 1], [1, 2], [1, 3]], "a:int,b:int")
    res = even_repartition(spark_session, df, 0, ["a"]).rdd.mapPartitions(_pc).collect()
    assert 7 == len(res)
    assert 4 == len([x for x in res if x[2] == 4])
    assert 3 == len([x for x in res if x[2] == 3])
    res = (
        even_repartition(spark_session, df, 0, ["a", "b"])
        .rdd.mapPartitions(_pc)
        .collect()
    )
    assert 7 == len(res)
    assert 7 == len([x for x in res if x[2] == 1])
    res = (
        even_repartition(spark_session, df, 1, ["a", "b"])
        .rdd.mapPartitions(_pc)
        .collect()
    )
    assert 7 == len(res)
    assert 7 == len([x for x in res if x[2] == 7])
    res = even_repartition(spark_session, df, 3, ["a"]).rdd.mapPartitions(_pc).collect()
    assert 7 == len(res)
    assert 4 == len([x for x in res if x[2] == 4])
    assert 3 == len([x for x in res if x[2] == 3])
    res = even_repartition(spark_session, df, 0, ["b"]).rdd.mapPartitions(_pc).collect()
    assert 7 == len(res)
    assert 6 == len([x for x in res if x[2] == 2])
    assert 1 == len([x for x in res if x[2] == 1])

    # test with multiple keys and that are not the first positions
    df = _df(
        [
            [1, "a", 1],
            [2, "b", 2],
            [3, "c", 3],
            [4, "d", 4],
            [5, "e", 1],
            [6, "f", 2],
            [7, "g", 3],
        ],
        "z:int,a:str,b:int",
    )
    res = (
        even_repartition(spark_session, df, 0, ["b", "z"])
        .rdd.mapPartitions(_pc)
        .collect()
    )
    assert sorted(res) == sorted(
        [
            [1, "a", 1, 1],
            [2, "b", 2, 1],
            [3, "c", 3, 1],
            [4, "d", 4, 1],
            [5, "e", 1, 1],
            [6, "f", 2, 1],
            [7, "g", 3, 1],
        ]
    )


def _pc(df):
    data = list(df)
    for row in data:
        yield list(row) + [len(data)]


def _df(data, schema=None):
    session = SparkSession.builder.getOrCreate()
    if schema is not None:
        pdf = PandasDataFrame(data, to_schema(schema))
        return session.createDataFrame(pdf.native, to_spark_schema(schema))
    else:
        return session.createDataFrame(data)



import os
from typing import Any, List

import duckdb
import pandas as pd
import ray
import ray.data as rd
from triad import FileSystem

import fugue.api as fa
from fugue import ArrayDataFrame, DataFrame, FugueWorkflow, fsql, transform
from fugue.dataframe.utils import _df_eq as df_eq
from fugue.plugins import infer_execution_engine
from fugue_ray import RayDataFrame, RayExecutionEngine
from fugue_test.builtin_suite import BuiltInTests
from fugue_test.execution_suite import ExecutionEngineTests

_CONF = {
    "fugue.rpc.server": "fugue.rpc.flask.FlaskRPCServer",
    "fugue.rpc.flask_server.host": "127.0.0.1",
    "fugue.rpc.flask_server.port": "1234",
    "fugue.rpc.flask_server.timeout": "2 sec",
}


class RayExecutionEngineTests(ExecutionEngineTests.Tests):
    @classmethod
    def setUpClass(cls):
        ray.init(num_cpus=2)
        cls._con = duckdb.connect()
        cls._engine = cls.make_engine(cls)
        fa.set_global_engine(cls._engine)

    @classmethod
    def tearDownClass(cls):
        fa.clear_global_engine()
        cls._con.close()
        ray.shutdown()

    def make_engine(self):
        e = RayExecutionEngine(
            conf={
                "test": True,
                "fugue.duckdb.pragma.threads": 2,
                "fugue.ray.zero_copy": True,
                "fugue.ray.default.batch_size": 10000,
            },
            connection=self._con,
        )
        return e

    def test_properties(self):
        assert self.engine.is_distributed
        assert self.engine.map_engine.is_distributed
        assert not self.engine.sql_engine.is_distributed

    def test_get_parallelism(self):
        assert fa.get_current_parallelism() == 2

    def test_repartitioning(self):
        # schema: *
        def t(df: pd.DataFrame) -> pd.DataFrame:
            return df.head(1)

        test = pd.DataFrame(dict(a=[1, 2, 3]))

        # even partitioning
        res = transform(
            test,
            t,
            partition="per_row",
            engine="ray",
            as_local=True,
            as_fugue=True,
        )
        df_eq(
            res,
            ArrayDataFrame([[1], [2], [3]], schema="a:long"),
            check_order=False,
            throw=True,
        )

        # rand partitioning
        res = transform(
            test,
            t,
            partition=dict(num=3, algo="rand"),
            engine="ray",
            as_local=True,
            as_fugue=True,
        )
        df_eq(
            res,
            ArrayDataFrame([[1], [2], [3]], schema="a:long"),
            check_order=False,
            throw=True,
        )

        # over partitioning
        res = transform(
            test,
            t,
            partition=dict(num=40),
            engine="ray",
            as_local=True,
            as_fugue=True,
        )
        df_eq(
            res,
            ArrayDataFrame([[1], [2], [3]], schema="a:long"),
            check_order=False,
            throw=True,
        )

    def test_partition_by_more(self):
        # schema: *
        def t(df: pd.DataFrame) -> pd.DataFrame:
            return df.head(1)

        test = pd.DataFrame(dict(a=[1, 2, 3]))

        res = transform(
            test,
            t,
            partition={"by": "a"},
            engine="ray",
            engine_conf={
                "fugue.ray.shuffle.partitions": 2,
                "fugue.ray.remote.num_cpus": 1,
            },
            as_local=True,
            as_fugue=True,
        )
        df_eq(
            res,
            ArrayDataFrame([[1], [2], [3]], schema="a:long"),
            check_order=False,
            throw=True,
        )

    def test_load_parquet_more(self):
        pq = os.path.join(str(self.tmpdir), "tmp.parquet")
        pd.DataFrame(dict(a=[1, 2], b=[3, 4], c=["a", "b"])).to_parquet(pq)
        tdf = self.engine.load_df(pq, columns="b:int,a:str").as_array()
        assert tdf == [[3, "1"], [4, "2"]]

    def test_load_json_more(self):
        js = os.path.join(str(self.tmpdir), "tmp.json")
        pdf = pd.DataFrame(dict(a=[1, 2], b=[3, 4], c=["a", "b"]))
        self.engine.save_df(self.engine.to_df(pdf), js)
        tdf = self.engine.load_df(js).as_array()
        assert tdf == [[1, 3, "a"], [2, 4, "b"]]
        tdf = self.engine.load_df(js, columns="b:int,a:str").as_array()
        assert tdf == [[3, "1"], [4, "2"]]

    def test_remote_args(self):
        e = RayExecutionEngine(conf={"fugue.ray.remote.num_cpus": 3})
        assert e._get_remote_args() == {"num_cpus": 3}

        e = RayExecutionEngine(conf={"fugue.ray.remote.scheduling_strategy": "SPREAD"})
        assert e._get_remote_args() == {"scheduling_strategy": "SPREAD"}

    def test_infer_engine(self):
        df = rd.from_pandas(pd.DataFrame([[0]], columns=["a"]))
        assert infer_execution_engine([df]) == "ray"

        fdf = RayDataFrame(df)
        assert infer_execution_engine([fdf]) == "ray"


class RayBuiltInTests(BuiltInTests.Tests):
    @classmethod
    def setUpClass(cls):
        ray.init(num_cpus=2)
        cls._con = duckdb.connect()
        cls._engine = cls.make_engine(cls)

    @classmethod
    def tearDownClass(cls):
        cls._con.close()
        ray.shutdown()

    def make_engine(self):
        e = RayExecutionEngine(
            conf={"test": True, "fugue.duckdb.pragma.threads": 2, **_CONF},
            connection=self._con,
        )
        return e

    def test_yield_table(self):
        pass

    def test_yield_2(self):
        def assert_data(df: DataFrame) -> None:
            assert df.schema == "a:datetime,b:bytes,c:[long]"

        df = pd.DataFrame([[1, 2, 3]], columns=list("abc"))

        with FugueWorkflow() as dag:
            x = dag.df(df)
            result = dag.select("SELECT * FROM ", x)
            result.yield_dataframe_as("x")
        res = dag.run(self.engine)
        assert res["x"].as_array() == [[1, 2, 3]]

    def test_io(self):
        path = os.path.join(self.tmpdir, "a")
        path2 = os.path.join(self.tmpdir, "b.test.csv")
        path3 = os.path.join(self.tmpdir, "c.partition")
        with FugueWorkflow() as dag:
            b = dag.df([[6, 1], [2, 7]], "c:int,a:long")
            b.partition(num=3).save(path, fmt="parquet", single=True)
            b.save(path2, header=True)
        dag.run(self.engine)
        assert FileSystem().isfile(path)
        with FugueWorkflow() as dag:
            a = dag.load(path, fmt="parquet", columns=["a", "c"])
            a.assert_eq(dag.df([[1, 6], [7, 2]], "a:long,c:int"))
            a = dag.load(path2, header=True, columns="c:int,a:long")
            a.assert_eq(dag.df([[6, 1], [2, 7]], "c:int,a:long"))
        dag.run(self.engine)

        return
        # TODO: the following (writing partitions) is not supported by Ray
        # with FugueWorkflow() as dag:
        #     b = dag.df([[6, 1], [2, 7]], "c:int,a:long")
        #     b.partition(by="c").save(path3, fmt="parquet", single=False)
        # assert FileSystem().isdir(path3)
        # assert FileSystem().isdir(os.path.join(path3, "c=6"))
        # assert FileSystem().isdir(os.path.join(path3, "c=2"))
        # # TODO: in test below, once issue #288 is fixed, use dag.load
        # #  instead of pd.read_parquet
        # pd.testing.assert_frame_equal(
        #     pd.read_parquet(path3).sort_values("a").reset_index(drop=True),
        #     pd.DataFrame({"c": pd.Categorical([6, 2]), "a": [1, 7]}).reset_index(
        #         drop=True
        #     ),
        #     check_like=True,
        # )

    def test_coarse_partition(self):
        def verify_coarse_partition(df: pd.DataFrame) -> List[List[Any]]:
            ct = df.a.nunique()
            s = df.a * 1000 + df.b
            ordered = ((s - s.shift(1)).dropna() >= 0).all(axis=None)
            return [[ct, ordered]]

        def assert_(df: pd.DataFrame, rc: int, n: int, check_ordered: bool) -> None:
            if rc > 0:
                assert len(df) == rc
            assert df.ct.sum() == n
            if check_ordered:
                assert (df.ordered == True).all()

        gps = 100
        partition_num = 6
        df = pd.DataFrame(dict(a=list(range(gps)) * 10, b=range(gps * 10))).sample(
            frac=1.0
        )
        with FugueWorkflow() as dag:
            a = dag.df(df)
            c = a.partition(
                algo="coarse", by="a", presort="b", num=partition_num
            ).transform(verify_coarse_partition, schema="ct:int,ordered:bool")
            dag.output(
                c,
                using=assert_,
                params=dict(rc=partition_num, n=gps, check_ordered=True),
            )
        dag.run(self.engine)



from typing import Any, Dict
from triad import ParamDict

KEYWORD_ROWCOUNT = "ROWCOUNT"
KEYWORD_PARALLELISM = "CONCURRENCY"

FUGUE_ENTRYPOINT = "fugue.plugins"

FUGUE_SQL_DEFAULT_DIALECT = "spark"

FUGUE_CONF_WORKFLOW_CONCURRENCY = "fugue.workflow.concurrency"
FUGUE_CONF_WORKFLOW_CHECKPOINT_PATH = "fugue.workflow.checkpoint.path"
FUGUE_CONF_WORKFLOW_AUTO_PERSIST = "fugue.workflow.auto_persist"
FUGUE_CONF_WORKFLOW_AUTO_PERSIST_VALUE = "fugue.workflow.auto_persist_value"
FUGUE_CONF_CACHE_PATH = "fugue.workflow.cache.path"
FUGUE_CONF_WORKFLOW_EXCEPTION_HIDE = "fugue.workflow.exception.hide"
FUGUE_CONF_WORKFLOW_EXCEPTION_INJECT = "fugue.workflow.exception.inject"
FUGUE_CONF_WORKFLOW_EXCEPTION_OPTIMIZE = "fugue.workflow.exception.optimize"
FUGUE_CONF_SQL_IGNORE_CASE = "fugue.sql.compile.ignore_case"
FUGUE_CONF_SQL_DIALECT = "fugue.sql.compile.dialect"
FUGUE_CONF_DEFAULT_PARTITIONS = "fugue.default.partitions"

FUGUE_COMPILE_TIME_CONFIGS = set(
    [
        FUGUE_CONF_WORKFLOW_AUTO_PERSIST,
        FUGUE_CONF_WORKFLOW_AUTO_PERSIST_VALUE,
        FUGUE_CONF_WORKFLOW_EXCEPTION_HIDE,
        FUGUE_CONF_WORKFLOW_EXCEPTION_INJECT,
        FUGUE_CONF_WORKFLOW_EXCEPTION_OPTIMIZE,
        FUGUE_CONF_SQL_IGNORE_CASE,
        FUGUE_CONF_SQL_DIALECT,
    ]
)

_FUGUE_GLOBAL_CONF = ParamDict(
    {
        FUGUE_CONF_WORKFLOW_CONCURRENCY: 1,
        FUGUE_CONF_WORKFLOW_AUTO_PERSIST: False,
        FUGUE_CONF_WORKFLOW_EXCEPTION_HIDE: "fugue.,six,adagio.,pandas,"
        "fugue_dask.,dask.,fugue_spark.,pyspark.,antlr4,_qpd_antlr,qpd,triad,"
        "fugue_notebook.,ipython.,jupyter.,ipykernel,_pytest,pytest,fugue_ibis.",
        FUGUE_CONF_WORKFLOW_EXCEPTION_INJECT: 3,
        FUGUE_CONF_WORKFLOW_EXCEPTION_OPTIMIZE: True,
        FUGUE_CONF_SQL_IGNORE_CASE: False,
        FUGUE_CONF_SQL_DIALECT: FUGUE_SQL_DEFAULT_DIALECT,
        FUGUE_CONF_DEFAULT_PARTITIONS: -1,
    }
)


def register_global_conf(
    conf: Dict[str, Any], on_dup: int = ParamDict.OVERWRITE
) -> None:
    """Register global Fugue configs that can be picked up by any
    Fugue execution engines as the base configs.

    :param conf: the config dictionary
    :param on_dup: see :meth:`triad.collections.dict.ParamDict.update`
      , defaults to ``ParamDict.OVERWRITE``

    .. note::

        When using ``ParamDict.THROW`` or ``on_dup``, it's transactional.
        If any key in ``conf`` is already in global config and the value
        is different from the new value, then ValueError will be thrown.

    .. admonition:: Examples

        .. code-block:: python

            from fugue import register_global_conf, NativeExecutionEngine

            register_global_conf({"my.value",1})

            engine = NativeExecutionEngine()
            assert 1 == engine.conf["my.value"]

            engine = NativeExecutionEngine({"my.value",2})
            assert 2 == engine.conf["my.value"]
    """
    if on_dup == ParamDict.THROW:
        # be transactional
        for k, v in conf.items():
            if k in _FUGUE_GLOBAL_CONF:
                vv = _FUGUE_GLOBAL_CONF[k]
                if vv != v:
                    raise ValueError(
                        f"for global config {k}, the existed "
                        f"value is {vv} and can't take new value {v}"
                    )
        on_dup = ParamDict.OVERWRITE
    _FUGUE_GLOBAL_CONF.update(conf, on_dup=on_dup, deep=True)



from abc import abstractmethod
from typing import Any, List, Optional

from ..dataset import Dataset, DatasetDisplay, get_dataset_display


class Bag(Dataset):
    """The base class of Fugue Bags. Bag contains a collection of
    unordered objects.
    """

    def as_local(self) -> "LocalBag":  # pragma: no cover
        """Convert this bag to a :class:`.LocalBag`"""
        return self.as_local_bounded()

    @abstractmethod
    def as_local_bounded(self) -> "LocalBoundedBag":  # pragma: no cover
        """Convert this bag to a :class:`.LocalBoundedBag`"""
        raise NotImplementedError

    @abstractmethod
    def peek(self) -> Any:  # pragma: no cover
        """Peek the first row of the dataframe as array

        :raises FugueDatasetEmptyError: if it is empty
        """
        raise NotImplementedError

    @abstractmethod
    def as_array(self) -> List[Any]:  # pragma: no cover
        """Convert to a native python array

        :return: the native python array
        """
        raise NotImplementedError

    @abstractmethod
    def head(self, n: int) -> "LocalBoundedBag":  # pragma: no cover
        """Take the first n elements

        :return: the python array of the first n elements
        """
        raise NotImplementedError

    def __copy__(self) -> "Bag":
        return self

    def __deepcopy__(self, memo: Any) -> "Bag":
        return self


class LocalBag(Bag):
    @property
    def is_local(self) -> bool:
        return True

    @property
    def num_partitions(self) -> int:
        return 1


class LocalBoundedBag(LocalBag):
    @property
    def is_bounded(self) -> bool:
        return True

    def as_local_bounded(self) -> "LocalBoundedBag":
        return self


class BagDisplay(DatasetDisplay):
    """:class:`~.Bag` plain display class"""

    @property
    def bg(self) -> Bag:
        """The target :class:`~.Bag`"""
        return self._ds  # type: ignore

    def show(
        self, n: int = 10, with_count: bool = False, title: Optional[str] = None
    ) -> None:
        head_rows = self.bg.head(n).as_array()
        if len(head_rows) < n:
            count = len(head_rows)
        else:
            count = self.bg.count() if with_count else -1
        with DatasetDisplay._SHOW_LOCK:
            if title is not None and title != "":
                print(title)
            print(type(self.bg).__name__)
            print(head_rows)
            if count >= 0:
                print(f"Total count: {count}")
                print("")
            if self.bg.has_metadata:
                print("Metadata:")
                try:
                    # try pretty print, but if not convertible to json, print original
                    print(self.bg.metadata.to_json(indent=True))
                except Exception:  # pragma: no cover
                    print(self.bg.metadata)
                print("")


@get_dataset_display.candidate(lambda ds: isinstance(ds, Bag), priority=0.1)
def _get_bag_display(ds: Bag):
    return BagDisplay(ds)



import os

import pandas as pd
from pytest import raises

from fugue import (
    ArrayDataFrame,
    DataFrame,
    FugueSQLWorkflow,
    NativeExecutionEngine,
    fsql,
    register_global_conf,
)
from fugue.dataframe.utils import _df_eq
from fugue.exceptions import FugueSQLError


def test_workflow_conf():
    dag = FugueSQLWorkflow({"x": 10})
    assert 10 == dag.conf.get_or_throw("x", int)
    assert not dag.conf.get_or_throw("fugue.sql.compile.ignore_case", bool)

    dag = FugueSQLWorkflow({"x": 10, "fugue.sql.compile.ignore_case": True})
    assert 10 == dag.conf.get_or_throw("x", int)
    assert dag.conf.get_or_throw("fugue.sql.compile.ignore_case", bool)

    dag = FugueSQLWorkflow({"fugue.sql.compile.ignore_case": "true"})
    assert dag.conf.get_or_throw("fugue.sql.compile.ignore_case", bool)


def test_conf_override():
    with raises(SyntaxError):
        FugueSQLWorkflow()("create [[0]] schema a:int")
    with FugueSQLWorkflow({"fugue.sql.compile.ignore_case": "true"}) as dag:
        a = dag.df([[0], [1]], "a:int")
        dag(
            """
        create [[0],[1]] schema a:int
        b = select *
        output a,b using assert_eq"""
        )
    dag.run(conf={"fugue.sql.compile.ignore_case": "true"})


def test_show():
    class _CustomShow(object):
        def __init__(self):
            self.called = False

        def show(self, schema, head_rows, title, rows, count):
            print(schema, head_rows)
            print(title, rows, count)
            self.called = True

    cs = _CustomShow()
    with FugueSQLWorkflow() as dag:
        dag(
            """
        a = CREATE[[0], [1]] SCHEMA a: int
        b = CREATE[[0], [1]] SCHEMA a: int
        PRINT 10 ROWS FROM a, b ROWCOUNT TITLE "abc"
        PRINT a, b
        """
        )
    dag.run()
    assert not cs.called
    with FugueSQLWorkflow() as dag:
        dag(
            """
        a = CREATE[[0], [1]] SCHEMA a: int
        b = CREATE[[0], [1]] SCHEMA a: int
        PRINT 10 ROWS FROM a, b ROWCOUNT TITLE "abc"
        PRINT a, b
        """
        )
    dag.run()


def test_builtin_plot():
    fsql(
        """
    CREATE [[0]] SCHEMA a:int
    OUTPUT USING viz:plot
    OUTPUT USING viz:hist(title="x")
    """
    ).run()

    fsql(
        """
    CREATE [[0,1],[2,0]] SCHEMA x:int,y:int
    OUTPUT PREPARTITION BY x PRESORT y USING viz:plot
    OUTPUT PREPARTITION BY x PRESORT y USING viz:plot(title="x")
    """
    ).run()


def test_jinja_keyword_in_sql():
    with FugueSQLWorkflow() as dag:
        dag(
            """{% raw -%}
            CREATE [["{%'{%'"]] SCHEMA a:str
            SELECT * WHERE a LIKE '{%'
            PRINT
            {%- endraw %}"""
        )

        df = dag.df([["b"]], "a:str")
        x = "b"
        dag(
            """
        df2 = SELECT *
        FROM df
        WHERE a = "{{x}}"
        OUTPUT df, df2 USING assert_eq
        """
        )
    dag.run(("", "duckdb"))


def test_use_df(tmpdir):
    # df generated inside dag
    with FugueSQLWorkflow() as dag:
        a = dag.df([[0], [1]], "a:int")
        dag(
            """
        b=CREATE[[0], [1]] SCHEMA a: int
        OUTPUT a, b USING assert_eq
        """
        )
        dag.sql_vars["b"].assert_eq(a)
    dag.run()
    # external non-workflowdataframe
    arr = ArrayDataFrame([[0], [1]], "a:int")
    with FugueSQLWorkflow() as dag:
        dag(
            """
        b=CREATE[[0], [1]] SCHEMA a: int
        OUTPUT a, b USING assert_eq
        """,
            a=arr,
        )
        dag.sql_vars["b"].assert_eq(dag.df([[0], [1]], "a:int"))
    dag.run()
    # from yield file
    engine = NativeExecutionEngine(
        conf={"fugue.workflow.checkpoint.path": os.path.join(tmpdir, "ck")}
    )
    with FugueSQLWorkflow() as dag:
        dag("CREATE[[0], [1]] SCHEMA a: int YIELD FILE AS b")
        res = dag.yields["b"]
    dag.run(engine)
    with FugueSQLWorkflow() as dag:
        dag(
            """
        b=CREATE[[0], [1]] SCHEMA a: int
        OUTPUT a, b USING assert_eq
        """,
            a=res,
        )
    dag.run(engine)

    # from yield dataframe
    engine = NativeExecutionEngine()
    with FugueSQLWorkflow() as dag:
        dag("CREATE[[0], [1]] SCHEMA a: int YIELD DATAFRAME AS b")
        res = dag.yields["b"]
    dag.run(engine)

    with FugueSQLWorkflow() as dag:
        dag(
            """
        b=CREATE[[0], [1]] SCHEMA a: int
        OUTPUT a, b USING assert_eq
        """,
            a=res,
        )
    dag.run(engine)


def test_use_soecial_df(tmpdir):
    # external non-workflowdataframe
    arr = ArrayDataFrame([[0], [1]], "a:int")
    fsql(
        """
        b=CREATE[[0], [1]] SCHEMA a: int
        a = SELECT * FROM a.x
        OUTPUT a, b USING assert_eq
        a = SELECT x.* FROM a.x AS x
        OUTPUT a, b USING assert_eq
        c=CREATE [[0,0],[1,1]] SCHEMA a:int,b:int
        d = SELECT x.*,y.a AS b FROM a.x x INNER JOIN a.x y ON x.a=y.a
        OUTPUT c, d USING assert_eq
        """,
        {"a.x": arr},
    ).run()

    # from yield file
    engine = NativeExecutionEngine(
        conf={"fugue.workflow.checkpoint.path": os.path.join(tmpdir, "ck")}
    )
    with FugueSQLWorkflow() as dag:
        dag("CREATE[[0], [1]] SCHEMA a: int YIELD FILE AS b")
        res = dag.yields["b"]
    dag.run(engine)
    with FugueSQLWorkflow() as dag:
        dag(
            """
        b=CREATE[[0], [1]] SCHEMA a: int
        a = SELECT * FROM a.x
        OUTPUT a, b USING assert_eq
        """,
            {"a.x": res},
        )
    dag.run(engine)


def test_lazy_use_df():
    df1 = pd.DataFrame([[0]], columns=["a"])
    df2 = pd.DataFrame([[1]], columns=["a"])
    # although df2 is defined as a local variable
    # since it is not used in dag1, so it was never converted

    dag1 = FugueSQLWorkflow()
    dag1("""PRINT df1""")

    dag2 = FugueSQLWorkflow()
    dag2.df(df1).show()

    assert dag1.spec_uuid() == dag2.spec_uuid()


def test_use_param():
    with FugueSQLWorkflow() as dag:
        a = dag.df([[0], [1]], "a:int")
        x = 0
        dag(
            """
        b=CREATE[[{{x}}], [{{y}}]] SCHEMA a: int
        OUTPUT a, b USING assert_eq
        """,
            y=1,
        )
    dag.run()


def test_multiple_sql():
    with FugueSQLWorkflow() as dag:
        a = dag.df([[0], [1]], "a:int")
        dag("b = CREATE [[0],[1]] SCHEMA a:int")
        dag("OUTPUT a,b USING assert_eq")
    dag.run()


def test_multiple_sql_with_reset():
    with FugueSQLWorkflow() as dag:
        a = dag.df([[0], [1]], "a:int")
        dag("b = CREATE [[0],[1]] SCHEMA a:int")
        a = dag.df([[0], [2]], "a:int")
        b = dag.df([[0], [2]], "a:int")
        dag(
            """
        OUTPUT a, b USING assert_eq
        OUTPUT a, (CREATE[[0], [2]] SCHEMA a: int) USING assert_eq
        """
        )
    dag.run()


def test_multiple_blocks():
    with FugueSQLWorkflow() as dag:
        a = dag.df([[0], [1]], "a:int")
        c = 1
        dag(
            """
        OUTPUT a, (CREATE[[0], [1]] SCHEMA a: int) USING assert_eq
        """
        )
    dag.run()
    # dataframe can't pass to another workflow
    with FugueSQLWorkflow() as dag:
        assert "a" in locals()
        with raises(FugueSQLError):
            dag(
                """
            OUTPUT a, (CREATE[[0], [1]] SCHEMA a: int) USING assert_eq
            """
            )
    dag.run()
    # other local variables are fine
    with FugueSQLWorkflow() as dag:
        a = dag.df([[0], [1]], "a:int")
        dag(
            """
        OUTPUT a, (CREATE[[0], [{{c}}]] SCHEMA a: int) USING assert_eq
        """
        )
    dag.run()


def test_function_calls():
    with FugueSQLWorkflow() as dag:
        a = dag.df([[0], [1]], "a:int")
        _eq(dag, a)
    dag.run()


def test_local_instance_as_extension():
    class _Mock(object):
        # schema: *
        def t(self, df: pd.DataFrame) -> pd.DataFrame:
            return df

        def test(self):
            with FugueSQLWorkflow() as dag:
                dag(
                    """
                a = CREATE [[0],[1]] SCHEMA a:int
                b = TRANSFORM USING self.t
                OUTPUT a,b USING assert_eq
                """
                )
            dag.run()

    m = _Mock()
    m.test()
    with FugueSQLWorkflow() as dag:
        dag(
            """
        a = CREATE [[0],[1]] SCHEMA a:int
        b = TRANSFORM USING m.t
        OUTPUT a,b USING assert_eq
        """
        )
    dag.run()


def test_call_back():
    class CB(object):
        def __init__(self):
            self.n = 0

        def incr(self, n):
            self.n += n
            return self.n

    cb = CB()

    # schema: *
    def t(df: pd.DataFrame, incr: callable) -> pd.DataFrame:
        incr(1)
        return df

    with FugueSQLWorkflow() as dag:
        dag(
            """
        a = CREATE [[0],[1],[1]] SCHEMA a:int
        TRANSFORM PREPARTITION BY a USING t CALLBACK cb.incr PERSIST
        OUTTRANSFORM a PREPARTITION BY a USING t CALLBACK cb.incr
        """
        )
    dag.run()

    assert 4 == cb.n


def test_fsql():
    # schema: *,x:long
    def t(df: pd.DataFrame) -> pd.DataFrame:
        df["x"] = 1
        return df

    df = pd.DataFrame([[0], [1]], columns=["a"])
    result = fsql(
        """
    SELECT * FROM df WHERE `a`>{{p}}
    UNION ALL
    SELECT * FROM df2 WHERE a>{{p}}
    TRANSFORM USING t YIELD DATAFRAME AS result
    """,
        df2=pd.DataFrame([[0], [1]], columns=["a"]),
        p=0,
    ).run()
    assert [[1, 1], [1, 1]] == result["result"].as_array()

    result = fsql(
        """
    select * from df where a>{{p}}
    union all
    select * from df2 where a>{{p}}
    transform using t yield dataframe as result
    """,
        df2=pd.DataFrame([[0], [1]], columns=["a"]),
        p=0,
        fsql_ignore_case=True,
    ).run()
    assert [[1, 1], [1, 1]] == result["result"].as_array()

    result = fsql(
        """
    SELECT * FROM df WHERE "a">{{p}}
    UNION ALL
    SELECT * FROM df2 WHERE a>{{p}}
    TRANSFORM USING t YIELD DATAFRAME AS result
    """,
        df2=pd.DataFrame([[0], [1]], columns=["a"]),
        p=0,
        fsql_dialect="duckdb",
    ).run()
    assert [[1, 1], [1, 1]] == result["result"].as_array()

    register_global_conf({"fugue.sql.compile.dialect": "duckdb"})
    try:
        result = fsql(
            """
        SELECT * FROM df WHERE "a">{{p}}
        UNION ALL
        SELECT * FROM df2 WHERE a>{{p}}
        TRANSFORM USING t YIELD DATAFRAME AS result
        """,
            df2=pd.DataFrame([[0], [1]], columns=["a"]),
            p=0,
        ).run()
        assert [[1, 1], [1, 1]] == result["result"].as_array()
    finally:
        register_global_conf({"fugue.sql.compile.dialect": "spark"})


def test_fsql_syntax_error():
    with raises(SyntaxError):
        fsql("""CREATEE [[0]] SCHEMA a:int""")


def _eq(dag, a):
    dag(
        """
    OUTPUT a, (CREATE[[0], [1]] SCHEMA a: int) USING assert_eq
    """
    )


def assert_eq(df1: DataFrame, df2: DataFrame) -> None:
    _df_eq(df1, df2, throw=True)



from typing import Any, Dict, Iterable, List

import pandas as pd
from fugue.dataframe import ArrayDataFrame, DataFrame, DataFrames
from fugue.exceptions import FugueInterfacelessError
from fugue.execution import ExecutionEngine, NativeExecutionEngine
from fugue.extensions.processor import (
    Processor,
    _to_processor,
    parse_processor,
    processor,
    register_processor,
)
from pytest import raises
from triad.collections.dict import ParamDict
from triad.utils.hash import to_uuid


def test_processor():
    assert isinstance(t1, Processor)
    assert isinstance(t2, Processor)


def test_register():
    register_processor("x", MockProcessor)
    b = _to_processor("x")
    assert isinstance(b, MockProcessor)

    raises(
        KeyError,
        lambda: register_processor("x", MockProcessor, on_dup=ParamDict.THROW),
    )


def test__to_processor():
    a = _to_processor(MockProcessor)
    assert isinstance(a, MockProcessor)
    b = _to_processor("MockProcessor")
    assert isinstance(b, MockProcessor)

    a = _to_processor(T0)
    assert isinstance(a, Processor)
    a = _to_processor(T0())

    assert isinstance(a, Processor)
    a = _to_processor(t1)
    assert isinstance(a, Processor)
    a._x = 1
    b = _to_processor(t1)
    assert isinstance(b, Processor)
    assert "_x" not in b.__dict__
    c = _to_processor(t1)
    assert isinstance(c, Processor)
    assert "_x" not in c.__dict__
    c._x = 1
    d = _to_processor("t1")
    assert isinstance(d, Processor)
    assert "_x" not in d.__dict__
    raises(FugueInterfacelessError, lambda: _to_processor("abc"))

    assert isinstance(_to_processor(t3), Processor)
    assert isinstance(_to_processor(t4), Processor)
    assert isinstance(_to_processor(t5), Processor)
    assert isinstance(_to_processor(t6), Processor)
    raises(FugueInterfacelessError, lambda: _to_processor(t6, "a:int"))
    assert isinstance(_to_processor(t7, "a:int"), Processor)
    raises(FugueInterfacelessError, lambda: _to_processor(t7))
    assert isinstance(_to_processor(t8), Processor)
    assert isinstance(_to_processor(t9), Processor)
    assert isinstance(_to_processor(t10), Processor)
    assert isinstance(_to_processor(t11), Processor)


def test_run_processor():
    df = ArrayDataFrame([[0]], "a:int")
    dfs = DataFrames(df1=df, df2=df)
    dfs2 = DataFrames(df, df)
    assert not dfs2.has_key

    o1 = _to_processor(t3)
    assert 4 == o1(df, df, 2).as_array()[0][0]

    o1._params = ParamDict([("a", 2)], deep=False)
    o1._execution_engine = None
    assert 4 == o1.process(dfs).as_array()[0][0]
    o1._params = ParamDict([("a", 2)], deep=False)
    o1._execution_engine = None
    assert 4 == o1.process(dfs2).as_array()[0][0]

    o1 = _to_processor(t5)
    assert 4 == o1("dummy", dfs, 2)[0][0]
    assert 4 == o1("dummy", dfs2, 2)[0][0]
    o1._params = ParamDict([("a", 2)], deep=False)
    o1._execution_engine = NativeExecutionEngine()
    assert 4 == o1.process(dfs).as_array()[0][0]
    o1._params = ParamDict([("a", 2)], deep=False)
    o1._execution_engine = NativeExecutionEngine()
    assert 4 == o1.process(dfs2).as_array()[0][0]


def test__to_processor_determinism():
    a = _to_processor(t1, None)
    b = _to_processor(t1, None)
    c = _to_processor("t1", None)
    d = _to_processor("t2", None)
    assert a is not b
    assert to_uuid(a) == to_uuid(b)
    assert a is not c
    assert to_uuid(a) == to_uuid(c)
    assert to_uuid(a) != to_uuid(d)

    a = _to_processor(MockProcessor)
    b = _to_processor("MockProcessor")
    assert a is not b
    assert to_uuid(a) == to_uuid(b)


def test_to_processor_validation():
    @processor("b:int", input_has=" a , b ")
    def pv1(df: Iterable[Dict[str, Any]]) -> Iterable[Dict[str, Any]]:
        for r in df:
            r["b"] = 1
            yield r

    # input_has: a , b
    # schema: b:int
    def pv2(df: Iterable[Dict[str, Any]]) -> Iterable[Dict[str, Any]]:
        for r in df:
            r["b"] = 1
            yield r

    class MockProcessorV(Processor):
        @property
        def validation_rules(self):
            return {"input_is": "a:int,b:int"}

        def process(self, dfs):
            return dfs[0]

    a = _to_processor(pv1, None)
    assert {"input_has": ["a", "b"]} == a.validation_rules
    b = _to_processor(pv2, None)
    assert {"input_has": ["a", "b"]} == b.validation_rules
    c = _to_processor(MockProcessorV)
    assert {"input_is": "a:int,b:int"} == c.validation_rules


def test_parse_processor():
    class _C(Processor):
        def __init__(self, x):
            self._x = x

        def process(self, dfs):
            raise NotImplementedError

        def __uuid__(self) -> str:
            return to_uuid(super().__uuid__(), self._x)

    @parse_processor.candidate(lambda x: isinstance(x, str) and x.startswith("(("))
    def _parse(obj):
        return _C(obj)

    a = _to_processor("((abc")
    b = _to_processor("((bc")
    c = _to_processor("((abc")

    assert isinstance(a, _C)
    assert isinstance(b, _C)
    assert isinstance(c, _C)
    assert to_uuid(a) == to_uuid(c)
    assert to_uuid(a) != to_uuid(b)


class T0(Processor):
    def process(self, dfs) -> DataFrame:
        return dfs[0]


@processor("")
def t1(df: Iterable[Dict[str, Any]]) -> DataFrame:
    pass


@processor("a:int")
def t2(e: ExecutionEngine, df1: DataFrame, df2: DataFrame) -> List[List[Any]]:
    pass


def t3(df1: DataFrame, df2: DataFrame, a) -> DataFrame:
    value = df1.count() + df2.count() + a
    return ArrayDataFrame([[value]], "a:int")


def t4(e: ExecutionEngine, df1: DataFrame, df2: DataFrame, a: int, b: str) -> DataFrame:
    pass


@processor("a:int")
def t5(e: ExecutionEngine, dfs: DataFrames, a) -> List[List[Any]]:
    assert e is not None
    value = sum(x.count() for x in dfs.values()) + a
    return ArrayDataFrame([[value]], "a:int").as_array()


def t6(dfs: DataFrames) -> DataFrame:
    pass


def t7(dfs: DataFrames) -> List[List[Any]]:
    pass


# schema: a:int
def t8(e: ExecutionEngine, df1: DataFrame, df2: DataFrame) -> List[List[Any]]:
    pass


def t9(e: ExecutionEngine, df1: DataFrame, df2: DataFrame) -> pd.DataFrame:
    pass


def t10(e: ExecutionEngine, df1: DataFrame, df2: DataFrame) -> Iterable[pd.DataFrame]:
    pass


def t11(
    e: ExecutionEngine, df1: DataFrame, df2: DataFrame, **kwargs
) -> Iterable[pd.DataFrame]:
    pass


class MockProcessor(Processor):
    def process(self, dfs):
        pass



from abc import ABC, abstractmethod

from fugue.dataframe import DataFrame
from fugue.extensions.context import ExtensionContext


class Creator(ExtensionContext, ABC):
    """The interface is to generate single DataFrame from `params`.
    For example reading data from file should be a type of Creator.
    Creator is task level extension, running on driver, and execution engine aware.

    To implement this class, you should not have ``__init__``, please directly implement
    the interface functions.

    .. note::

      Before implementing this class, do you really need to implement this
      interface? Do you know the interfaceless feature of Fugue? Implementing Creator
      is commonly unnecessary. You can choose the interfaceless approach which may
      decouple your code from Fugue.

    .. seealso::

      Please read :doc:`Creator Tutorial <tutorial:tutorials/extensions/creator>`
    """

    @abstractmethod
    def create(self) -> DataFrame:  # pragma: no cover
        """Create DataFrame on driver side

        .. note::

          * It runs on driver side
          * The output dataframe is not necessarily local, for example a SparkDataFrame
          * It is engine aware, you can put platform dependent code in it (for example
            native pyspark code) but by doing so your code may not be portable. If you
            only use the functions of the general ExecutionEngine interface, it's still
            portable.

        :return: result dataframe
        """
        raise NotImplementedError



import pytest

ibis = pytest.importorskip("ibis")
from pytest import raises

from fugue import FugueWorkflow, NativeExecutionEngine
from fugue_ibis import PandasIbisEngine, as_fugue, as_ibis, parse_ibis_engine, run_ibis


def test_parse_ibis_engine():
    e = NativeExecutionEngine()
    ie = PandasIbisEngine(e)
    assert isinstance(parse_ibis_engine(e, e), PandasIbisEngine)
    assert isinstance(parse_ibis_engine(ie, e), PandasIbisEngine)
    with raises(NotImplementedError):
        parse_ibis_engine("dummy", e)


def test_run_ibis():
    def _test1(con: ibis.BaseBackend) -> ibis.Expr:
        tb = con.table("a")
        return tb

    def _test2(con: ibis.BaseBackend) -> ibis.Expr:
        tb = con.table("a")
        return tb.mutate(c=tb.a + tb.b)

    with FugueWorkflow() as dag:
        df = dag.df([[0, 1], [2, 3]], "a:long,b:long")
        res = run_ibis(_test1, a=df)
        res.assert_eq(df)
        df = dag.df([[0, 1], [2, 3]], "a:long,b:long")
        res = run_ibis(_test2, a=df)
        df2 = dag.df([[0, 1, 1], [2, 3, 5]], "a:long,b:long,c:long")
        res.assert_eq(df2)
    dag.run()


def test_run_as_ibis():
    with FugueWorkflow() as dag:
        df = dag.df([[0, 1], [2, 3]], "a:long,b:long")
        idf = as_ibis(df)
        res = as_fugue(idf)
        res.assert_eq(df)
    dag.run()

    with FugueWorkflow() as dag:
        df1 = dag.df([[0, 1], [2, 3]], "a:long,b:long")
        df2 = dag.df([[0, ["x"]], [3, ["y"]]], "a:long,c:[str]")
        idf1 = as_ibis(df1)
        idf2 = as_ibis(df2)
        idf = idf1.inner_join(idf2, idf1.a == idf2.a)[idf1, idf2.c]
        res = as_fugue(idf)
        expected = dag.df([[0, 1, ["x"]]], "a:long,b:long,c:[str]")
        res.assert_eq(expected, check_order=True, check_schema=True)
    dag.run()

    with FugueWorkflow() as dag:
        idf1 = dag.df([[0, 1], [2, 3]], "a:long,b:long").as_ibis()
        idf2 = dag.df([[0, ["x"]], [3, ["y"]]], "a:long,c:[str]").as_ibis()
        res = idf1.inner_join(idf2, idf1.a == idf2.a)[idf1, idf2.c].as_fugue()
        expected = dag.df([[0, 1, ["x"]]], "a:long,b:long,c:[str]")
        res.assert_eq(expected, check_order=True, check_schema=True)
    dag.run()



# flake8: noqa
# TODO: This folder is to be deprecated
from fugue_version import __version__

import warnings
from fugue import FugueSQLWorkflow, fsql

warnings.warn(
    "fsql and FugueSQLWorkflow now should be imported directly from fugue, "
    "fugue_sql will be removed in 0.9.0"
)



import pandas as pd
from fugue import FugueWorkflow, WorkflowDataFrame, WorkflowDataFrames, module
from fugue.workflow.module import _to_module, _ModuleFunctionWrapper
from fugue import FugueSQLWorkflow
from pytest import raises
from fugue.exceptions import FugueInterfacelessError


def test_input_module():
    # pylint: disable=no-value-for-parameter
    @module
    def input1(wf: FugueWorkflow) -> WorkflowDataFrame:
        return wf.df([[0]], "a:int")

    @module()
    def input2(wf: FugueWorkflow, a: int) -> WorkflowDataFrame:
        return wf.df([[a]], "a:int")

    @module(as_method=True)
    def input3(wf: FugueWorkflow, a: int, b: int) -> WorkflowDataFrames:
        return WorkflowDataFrames(a=wf.df([[a]], "a:int"), b=wf.df([[b]], "b:int"))

    assert not input1.has_input
    assert input1.has_single_output
    assert not input1.has_no_output

    assert input3.has_multiple_output

    with FugueWorkflow() as dag:
        input1(dag).assert_eq(dag.df([[0]], "a:int"))
    dag.run()

    with FugueWorkflow() as dag:
        input2(a=10, wf=dag).assert_eq(dag.df([[10]], "a:int"))
    dag.run()

    with FugueWorkflow() as dag:
        dfs = input3(dag, 10, 11)
        dfs["a"].assert_eq(dag.df([[10]], "a:int"))
        dfs["b"].assert_eq(dag.df([[11]], "b:int"))
        dfs = dag.input3(10, 11)
        dfs["a"].assert_eq(dag.df([[10]], "a:int"))
        dfs["b"].assert_eq(dag.df([[11]], "b:int"))
    dag.run()


def test_process_module():
    # pylint: disable=no-value-for-parameter
    def process(df: pd.DataFrame, d: int = 1) -> pd.DataFrame:
        df["a"] += d
        return df

    @module
    def p1(wf: FugueSQLWorkflow, df: WorkflowDataFrame) -> WorkflowDataFrame:
        return df.process(process)

    @module()
    def p2(wf: FugueWorkflow, dfs: WorkflowDataFrames, d: int) -> WorkflowDataFrames:
        return WorkflowDataFrames(
            {k: v.process(process, params={"d": d}) for k, v in dfs.items()}
        )

    @module(as_method=True, name="p4")
    def p3(df: WorkflowDataFrame) -> WorkflowDataFrame:
        return df.process(process)

    assert p1.has_input
    assert not p1.has_dfs_input
    assert p2.has_dfs_input

    with FugueSQLWorkflow() as dag:
        df = dag.df([[0]], "a:int")
        p1(df).assert_eq(dag.df([[1]], "a:int"))
        p1(dag, df).assert_eq(dag.df([[1]], "a:int"))
        p1(df=df).assert_eq(dag.df([[1]], "a:int"))
        p1(df=df, wf=dag).assert_eq(dag.df([[1]], "a:int"))

    with FugueWorkflow() as dag:
        dfs = WorkflowDataFrames(aa=dag.df([[0]], "a:int"), bb=dag.df([[10]], "a:int"))
        r = p2(dag, dfs, 1)
        r["aa"].assert_eq(dag.df([[1]], "a:int"))
        r["bb"].assert_eq(dag.df([[11]], "a:int"))

        r = p2(dfs, 1)
        r["aa"].assert_eq(dag.df([[1]], "a:int"))
        r["bb"].assert_eq(dag.df([[11]], "a:int"))

        r = p2(d=1, dfs=dfs, wf=dag)
        r["aa"].assert_eq(dag.df([[1]], "a:int"))
        r["bb"].assert_eq(dag.df([[11]], "a:int"))

        r = p2(d=1, dfs=dfs)
        r["aa"].assert_eq(dag.df([[1]], "a:int"))
        r["bb"].assert_eq(dag.df([[11]], "a:int"))
    dag.run()

    with FugueWorkflow() as dag:
        df = dag.df([[0]], "a:int")
        p3(df).assert_eq(dag.df([[1]], "a:int"))
        p3(df=df).assert_eq(dag.df([[1]], "a:int"))
        df.p4().assert_eq(dag.df([[1]], "a:int"))
    dag.run()


def test_output_module():
    # pylint: disable=no-value-for-parameter

    @module
    def o1(wf: FugueWorkflow, df: WorkflowDataFrame) -> None:
        pass

    @module()
    def o2(wf: FugueWorkflow, df: WorkflowDataFrame):
        pass

    @module(as_method=True)
    def o3(df: WorkflowDataFrame):
        pass

    assert o1.has_input
    assert o1.has_no_output
    assert o2.has_no_output
    assert o3.has_no_output

    with FugueWorkflow() as dag:
        df = dag.df([[0]], "a:int")
        o1(df)
        o1(dag, df)
        o2(df=df)
        o2(df=df, wf=dag)
        o3(df)
        df.o3()
    dag.run()


def test_invalid_module():
    # pylint: disable=no-value-for-parameter

    @module()
    def o1(wf: FugueWorkflow, df1: WorkflowDataFrame, df2: WorkflowDataFrame) -> None:
        pass

    @module()
    def o2(wf: FugueWorkflow, dfs: WorkflowDataFrames) -> None:
        pass

    dag1 = FugueWorkflow()
    df1 = dag1.df([[0]], "a:int")
    dag2 = FugueWorkflow()
    df2 = dag2.df([[1]], "a:int")

    with raises(ValueError):
        o1(df1, df2)

    with raises(ValueError):
        o2(WorkflowDataFrames(a=df1, b=df2))


def test_interfaceless():
    def i1(wf: FugueSQLWorkflow) -> WorkflowDataFrame:
        return wf.df([[0]], "a:int")

    def i2(df: WorkflowDataFrame) -> WorkflowDataFrame:
        return df

    @module()
    def i3(wf: FugueSQLWorkflow, df: WorkflowDataFrame) -> WorkflowDataFrame:
        return df

    assert isinstance(_to_module(i1), _ModuleFunctionWrapper)
    assert isinstance(_to_module("i2"), _ModuleFunctionWrapper)
    assert isinstance(_to_module(i3), _ModuleFunctionWrapper)
    assert isinstance(_to_module("i3"), _ModuleFunctionWrapper)

    def i4():
        pass

    def i5(df: WorkflowDataFrame, wf: FugueWorkflow) -> None:
        pass

    raises(FugueInterfacelessError, lambda: _to_module(i4))
    raises(FugueInterfacelessError, lambda: _to_module(i5))



from typing import Any, Optional, Tuple

import pyspark.rdd as pr
import pyspark.sql as ps
from pyspark import SparkContext
from pyspark.sql import SparkSession
from triad import run_at_def

from fugue import DataFrame, ExecutionEngine, register_execution_engine
from fugue.dev import (
    DataFrameParam,
    ExecutionEngineParam,
    fugue_annotated_param,
    is_pandas_or,
)
from fugue.extensions import namespace_candidate
from fugue.plugins import as_fugue_dataset, infer_execution_engine, parse_creator
from fugue_spark.dataframe import SparkDataFrame
from fugue_spark.execution_engine import SparkExecutionEngine

_is_sparksql = namespace_candidate("sparksql", lambda x: isinstance(x, str))


@infer_execution_engine.candidate(
    lambda objs: is_pandas_or(objs, (ps.DataFrame, SparkDataFrame))
    or any(_is_sparksql(obj) for obj in objs)
)
def _infer_spark_client(obj: Any) -> Any:
    return SparkSession.builder.getOrCreate()


@as_fugue_dataset.candidate(lambda df, **kwargs: isinstance(df, ps.DataFrame))
def _spark_as_fugue_df(df: ps.DataFrame, **kwargs: Any) -> SparkDataFrame:
    return SparkDataFrame(df, **kwargs)


@parse_creator.candidate(_is_sparksql)
def _parse_sparksql_creator(sql: Tuple[str, str]):
    def _run_sql(spark: SparkSession) -> ps.DataFrame:
        return spark.sql(sql[1])

    return _run_sql


def _register_engines() -> None:
    register_execution_engine(
        "spark",
        lambda conf, **kwargs: SparkExecutionEngine(conf=conf),
        on_dup="ignore",
    )
    register_execution_engine(
        SparkSession,
        lambda session, conf, **kwargs: SparkExecutionEngine(session, conf=conf),
        on_dup="ignore",
    )


@fugue_annotated_param(SparkExecutionEngine)
class _SparkExecutionEngineParam(ExecutionEngineParam):
    pass


@fugue_annotated_param(SparkSession)
class _SparkSessionParam(ExecutionEngineParam):
    def to_input(self, engine: ExecutionEngine) -> Any:
        assert isinstance(engine, SparkExecutionEngine)
        return engine.spark_session  # type:ignore


@fugue_annotated_param(SparkContext)
class _SparkContextParam(ExecutionEngineParam):
    def to_input(self, engine: ExecutionEngine) -> Any:
        assert isinstance(engine, SparkExecutionEngine)
        return engine.spark_session.sparkContext  # type:ignore


@fugue_annotated_param(ps.DataFrame)
class _SparkDataFrameParam(DataFrameParam):
    def to_input_data(self, df: DataFrame, ctx: Any) -> Any:
        assert isinstance(ctx, SparkExecutionEngine)
        return ctx.to_df(df).native

    def to_output_df(self, output: Any, schema: Any, ctx: Any) -> DataFrame:
        assert isinstance(output, ps.DataFrame)
        assert isinstance(ctx, SparkExecutionEngine)
        return ctx.to_df(output, schema=schema)

    def count(self, df: Any) -> int:  # pragma: no cover
        raise NotImplementedError("not allowed")


@fugue_annotated_param(pr.RDD)
class _RddParam(DataFrameParam):
    def to_input_data(self, df: DataFrame, ctx: Any) -> Any:
        assert isinstance(ctx, SparkExecutionEngine)
        return ctx.to_df(df).native.rdd

    def to_output_df(self, output: Any, schema: Any, ctx: Any) -> DataFrame:
        assert isinstance(output, pr.RDD)
        assert isinstance(ctx, SparkExecutionEngine)
        return ctx.to_df(output, schema=schema)

    def count(self, df: Any) -> int:  # pragma: no cover
        raise NotImplementedError("not allowed")

    def need_schema(self) -> Optional[bool]:
        return True


@run_at_def
def _register() -> None:
    """Register Spark Execution Engine

    .. note::

        This function is automatically called when you do

        >>> import fugue_spark
    """
    _register_engines()



import os
from typing import Any, Dict, Iterable

import pandas as pd

from fugue import (
    DataFrame,
    FugueWorkflow,
    PandasDataFrame,
    out_transform,
    transform,
    make_execution_engine,
)
from fugue.constants import FUGUE_CONF_WORKFLOW_CHECKPOINT_PATH
from fugue.exceptions import FugueInterfacelessError, FugueWorkflowCompileError
from pytest import raises


def test_transform():
    pdf = pd.DataFrame([[1, 10], [0, 0], [1, 1], [0, 20]], columns=["a", "b"])

    def f1(df: pd.DataFrame) -> pd.DataFrame:
        return df.sort_values("b").head(1)

    result = transform(pdf, f1, schema="*")
    assert isinstance(result, pd.DataFrame)
    assert result.values.tolist() == [[0, 0]]

    # schema: *
    def f2(df: pd.DataFrame) -> pd.DataFrame:
        return df.sort_values("b").head(1)

    result = transform(pdf, f2)
    assert isinstance(result, pd.DataFrame)
    assert result.values.tolist() == [[0, 0]]

    result = transform(pdf, f2, persist=True)
    assert isinstance(result, pd.DataFrame)
    assert result.values.tolist() == [[0, 0]]

    result = transform(pdf, f2, partition=dict(by=["a"]))
    assert isinstance(result, pd.DataFrame)
    assert sorted(result.values.tolist(), key=lambda x: x[0]) == [[0, 0], [1, 1]]
    result = transform(
        pdf, f2, partition=dict(by=["a"]), as_fugue=True
    )
    assert isinstance(result, DataFrame)

    ppdf = PandasDataFrame(pdf)
    assert isinstance(transform(ppdf, f2), DataFrame)

    # schema: *
    def f3(df: pd.DataFrame, called: callable) -> pd.DataFrame:
        called()
        return df

    cb = Callback()
    result = transform(pdf, f3, callback=cb.called)
    assert 1 == cb.ct

    with raises(FugueWorkflowCompileError):
        transform(123, f2)


def test_transform_from_yield(tmpdir):
    # schema: *,x:int
    def f(df: pd.DataFrame) -> pd.DataFrame:
        return df.assign(x=1)

    dag = FugueWorkflow()
    dag.df([[0]], "a:int").yield_dataframe_as("x1")
    dag.df([[1]], "b:int").yield_file_as("x2")
    dag.run("", {FUGUE_CONF_WORKFLOW_CHECKPOINT_PATH: str(tmpdir)})

    result = transform(dag.yields["x1"], f)
    assert isinstance(result, DataFrame)
    assert result.as_array(type_safe=True) == [[0, 1]]

    result = transform(
        dag.yields["x2"],
        f,
        engine_conf={FUGUE_CONF_WORKFLOW_CHECKPOINT_PATH: str(tmpdir)},
    )
    assert isinstance(result, DataFrame)
    assert result.as_array(type_safe=True) == [[1, 1]]


def test_transform_from_file(tmpdir):
    fp = os.path.join(tmpdir, "t.parquet")
    pd.DataFrame(dict(a=[2])).to_parquet(fp)

    # schema: *,x:int
    def f(df: pd.DataFrame) -> pd.DataFrame:
        return df.assign(x=1)

    result = transform(fp, f, as_fugue=True)
    assert result.as_array(type_safe=True) == [[2, 1]]

    with raises(FugueInterfacelessError):
        transform("t.csv", f)

    with raises(FugueInterfacelessError):
        transform("t.json", f)


def test_transform_to_file(tmpdir):
    fp = os.path.join(tmpdir, "t.parquet")
    tdf = pd.DataFrame(dict(a=[2]))
    # schema: *,x:int
    def f(df: pd.DataFrame) -> pd.DataFrame:
        return df.assign(x=1)

    engine = make_execution_engine(
        None, {FUGUE_CONF_WORKFLOW_CHECKPOINT_PATH: str(tmpdir)}
    )

    # checkpoint is True, save_path is None
    result = transform(
        tdf, f, as_fugue=True, checkpoint=True, engine=engine
    )
    assert result.as_array() == [[2, 1]]

    # checkpoint is True, save_path is not None
    result = transform(
        tdf,
        f,
        as_fugue=True,
        checkpoint=True,
        save_path=fp,
        engine=engine,
    )
    assert result.as_array() == [[2, 1]]
    os.remove(fp)

    # checkpoint is False, save_path is not None
    result = transform(
        tdf,
        f,
        as_fugue=True,
        save_path=fp,
        engine=engine,
    )
    assert os.path.exists(fp)
    assert result == fp
    assert pd.read_parquet(fp).values.tolist() == [[2, 1]]
    os.remove(fp)

    # test that parquet format is used for saving when
    # no file extension is provided
    fp = os.path.join(tmpdir, "test")
    transform(
        tdf,
        f,
        save_path=fp,
        engine=engine,
    )
    loaded = pd.read_parquet(fp)
    assert pd.read_parquet(fp).values.tolist() == [[2, 1]]
    os.remove(fp)

    # catch invalid file paths
    with raises(FugueInterfacelessError):
        transform(
            tdf,
            f,
            as_fugue=True,
            save_path="f.csv",
            engine=engine,
        )
    with raises(FugueInterfacelessError):
        transform(
            tdf,
            f,
            as_fugue=True,
            save_path="f.json",
            engine=engine,
        )


def test_out_transform(tmpdir):
    pdf = pd.DataFrame([[1, 10], [0, 0], [1, 1], [0, 20]], columns=["a", "b"])

    class T:
        def __init__(self):
            self.n = 0

        def f(self, df: Iterable[Dict[str, Any]]) -> None:
            self.n += 1

    t = T()
    out_transform(pdf, t.f)
    assert 1 == t.n

    t = T()
    out_transform(pdf, t.f, partition=dict(by=["a"]))
    assert 2 == t.n

    dag = FugueWorkflow()
    dag.df(pdf).yield_dataframe_as("x1")
    dag.df(pdf).yield_dataframe_as("x2")
    dag.run("", {FUGUE_CONF_WORKFLOW_CHECKPOINT_PATH: str(tmpdir)})

    t = T()
    out_transform(dag.yields["x1"], t.f)
    assert 1 == t.n

    t = T()
    out_transform(
        dag.yields["x2"],
        t.f,
        partition=dict(by=["a"]),
        engine_conf={FUGUE_CONF_WORKFLOW_CHECKPOINT_PATH: str(tmpdir)},
    )
    assert 2 == t.n

    # schema: *
    def f3(df: pd.DataFrame, called: callable) -> pd.DataFrame:
        called()
        return df

    cb = Callback()
    out_transform(pdf, f3, callback=cb.called)
    assert 1 == cb.ct

    fp = os.path.join(tmpdir, "t.parquet")
    pdf.to_parquet(fp)
    cb = Callback()
    out_transform(fp, f3, callback=cb.called)
    assert 1 == cb.ct

    with raises(FugueWorkflowCompileError):
        out_transform(123, t.f)


class Callback:
    def __init__(self):
        self.ct = 0

    def called(self) -> None:
        self.ct += 1



# pylint: disable-all

from typing import (
    Any,
    Dict,
    Iterable,
    List,
    Optional,
    Tuple,
    Type,
    Union,
    no_type_check,
)

import pyarrow as pa
from antlr4 import ParserRuleContext
from antlr4.Token import CommonToken
from antlr4.tree.Tree import TerminalNode, Token, Tree
from fugue_sql_antlr import FugueSQLParser, FugueSQLVisitor
from fugue_sql_antlr._parser.fugue_sqlParser import fugue_sqlParser as fp
from triad import to_uuid
from triad.collections.schema import Schema
from triad.utils.assertion import assert_or_throw
from triad.utils.convert import (
    get_caller_global_local_vars,
    to_bool,
    to_function,
    to_type,
)
from triad.utils.pyarrow import to_pa_datatype
from triad.utils.schema import unquote_name, safe_split_out_of_quote
from triad.utils.string import validate_triad_var_name

from ..collections.partition import PartitionSpec
from ..exceptions import FugueSQLError, FugueSQLRuntimeError, FugueSQLSyntaxError
from ..execution.execution_engine import SQLEngine
from ..extensions.creator.convert import _to_creator
from ..extensions.outputter.convert import _to_outputter
from ..extensions.processor.convert import _to_processor
from ..extensions.transformer.convert import _to_output_transformer, _to_transformer
from ..workflow.module import _to_module
from ..workflow.workflow import FugueWorkflow, WorkflowDataFrame, WorkflowDataFrames
from ._utils import LazyWorkflowDataFrame


class FugueSQLHooks(object):
    def on_select_source_not_found(
        self, workflow: FugueWorkflow, name: str
    ) -> Union[WorkflowDataFrame, str]:
        return name


class _VisitorBase(FugueSQLVisitor):
    def __init__(self, sql: FugueSQLParser):
        super().__init__(sql)

    def visitFugueIdentifier(self, ctx: fp.FugueIdentifierContext) -> str:
        return self.ctxToStr(ctx)

    def collectChildren(self, node: Tree, tp: Type) -> List[Any]:
        result: List[Any] = []
        n = node.getChildCount()
        for i in range(n):
            c = node.getChild(i)
            if isinstance(c, tp):
                result.append(c.accept(self))
        return result

    def ctxToStr(self, node: Union[Tree, Token, None], delimit: str = " ") -> str:
        return self.sql.get_norm_text(node, delimiter=delimit)

    def to_runtime_error(self, ctx: ParserRuleContext) -> Exception:
        msg = "\n" + self.sql.get_raw_text(ctx, add_lineno=True)
        return FugueSQLRuntimeError(msg)

    @no_type_check
    def to_kv(self, ctx: Tree) -> Tuple[Any, Any]:
        k = ctx.key.accept(self)
        v = ctx.value.accept(self)
        return (k, v)

    def get_dict(self, ctx: Tree, *keys: Any) -> Dict[str, Any]:
        res: Dict[str, Any] = {}
        for k in keys:
            v = getattr(ctx, k)
            if v is not None:
                res[k] = self.visit(v)
        return res

    def visitFugueJsonObj(self, ctx: fp.FugueJsonObjContext) -> Any:
        pairs = ctx.fugueJsonPairs()
        if pairs is None:
            return dict()
        return pairs.accept(self)

    def visitFugueJsonPairs(self, ctx: fp.FugueJsonPairsContext) -> Dict:
        return dict(self.collectChildren(ctx, fp.FugueJsonPairContext))

    def visitFugueJsonPair(self, ctx: fp.FugueJsonPairContext) -> Any:
        return self.to_kv(ctx)

    def visitFugueJsonArray(self, ctx: fp.FugueJsonArrayContext) -> List[Any]:
        return self.collectChildren(ctx, fp.FugueJsonValueContext)

    def visitFugueJsonString(self, ctx: fp.FugueJsonKeyContext) -> Any:
        return eval(self.ctxToStr(ctx))

    def visitFugueJsonNumber(self, ctx: fp.FugueJsonKeyContext) -> Any:
        return eval(self.ctxToStr(ctx))

    def visitFugueJsonBool(self, ctx: fp.FugueJsonKeyContext) -> bool:
        return to_bool(self.ctxToStr(ctx))

    def visitFugueJsonNull(self, ctx: fp.FugueJsonKeyContext) -> Any:
        return None

    def visitFugueRenamePair(self, ctx: fp.FugueRenamePairContext) -> Tuple:
        return self.to_kv(ctx)

    def visitFugueRenameExpression(
        self, ctx: fp.FugueRenameExpressionContext
    ) -> Dict[str, str]:
        return dict(self.collectChildren(ctx, fp.FugueRenamePairContext))

    def visitFugueWildSchema(self, ctx: fp.FugueWildSchemaContext) -> str:
        schema = ",".join(self.collectChildren(ctx, fp.FugueWildSchemaPairContext))
        ops = "".join(self.collectChildren(ctx, fp.FugueSchemaOpContext))
        return schema + ops

    def visitFugueWildSchemaPair(self, ctx: fp.FugueWildSchemaPairContext) -> str:
        if ctx.pair is not None:
            return str(Schema([self.visit(ctx.pair)]))
        else:
            return "*"

    def visitFugueSchemaOp(self, ctx: fp.FugueSchemaOpContext) -> str:
        return self.ctxToStr(ctx, delimit="")

    def visitFugueSchema(self, ctx: fp.FugueSchemaContext) -> Schema:
        return Schema(self.collectChildren(ctx, fp.FugueSchemaPairContext))

    def visitFugueSchemaPair(self, ctx: fp.FugueSchemaPairContext) -> Any:
        tp = self.to_kv(ctx)
        return (unquote_name(tp[0]), tp[1])

    def visitFugueSchemaSimpleType(
        self, ctx: fp.FugueSchemaSimpleTypeContext
    ) -> pa.DataType:
        return to_pa_datatype(self.ctxToStr(ctx))

    def visitFugueSchemaListType(
        self, ctx: fp.FugueSchemaListTypeContext
    ) -> pa.DataType:
        tp = self.visit(ctx.fugueSchemaType())
        return pa.list_(tp)

    def visitFugueSchemaStructType(
        self, ctx: fp.FugueSchemaStructTypeContext
    ) -> pa.DataType:
        fields = self.visit(ctx.fugueSchema()).fields
        return pa.struct(fields)

    def visitFugueSchemaMapType(self, ctx: fp.FugueSchemaMapTypeContext) -> pa.DataType:
        tps = self.collectChildren(ctx, fp.FugueSchemaTypeContext)
        return pa.map_(tps[0], tps[1])

    def visitFuguePrepartition(self, ctx: fp.FuguePrepartitionContext) -> PartitionSpec:
        params = self.get_dict(ctx, "algo", "num", "by", "presort")
        return PartitionSpec(**params)

    def visitFuguePartitionAlgo(self, ctx: fp.FuguePartitionAlgoContext) -> str:
        return self.ctxToStr(ctx).lower()

    def visitFuguePartitionNum(self, ctx: fp.FuguePartitionNumContext) -> str:
        return self.ctxToStr(ctx, delimit="").upper()

    def visitFugueCols(self, ctx: fp.FugueColsContext) -> List[str]:
        return self.collectChildren(ctx, fp.FugueColumnIdentifierContext)

    def visitFugueColsSort(self, ctx: fp.FugueColsSortContext) -> str:
        return ",".join(self.collectChildren(ctx, fp.FugueColSortContext))

    def visitFugueColSort(self, ctx: fp.FugueColSortContext) -> str:
        return self.ctxToStr(ctx)

    def visitFugueColumnIdentifier(self, ctx: fp.FugueColumnIdentifierContext) -> str:
        return unquote_name(self.ctxToStr(ctx))

    def visitFugueParamsPairs(self, ctx: fp.FugueParamsPairsContext) -> Dict:
        return dict(self.collectChildren(ctx.pairs, fp.FugueJsonPairContext))

    def visitFugueParamsObj(self, ctx: fp.FugueParamsObjContext) -> Any:
        return self.visit(ctx.obj)

    def visitFugueExtension(self, ctx: fp.FugueExtensionContext) -> Any:
        s = self.ctxToStr(ctx, delimit="")
        if ctx.domain is None:
            return s
        p = safe_split_out_of_quote(s, ":", 1)
        return unquote_name(p[0]), p[1]

    def visitFugueSingleOutputExtensionCommon(
        self, ctx: fp.FugueSingleOutputExtensionCommonContext
    ) -> Dict[str, Any]:
        return self.get_dict(ctx, "fugueUsing", "params", "schema")

    def visitFugueSingleOutputExtensionCommonWild(
        self, ctx: fp.FugueSingleOutputExtensionCommonContext
    ) -> Dict[str, Any]:
        return self.get_dict(ctx, "fugueUsing", "params", "schema")

    def visitFugueAssignment(self, ctx: fp.FugueAssignmentContext) -> Tuple:
        varname = self.ctxToStr(ctx.varname, delimit="")
        sign = self.ctxToStr(ctx.sign, delimit="")
        return varname, sign

    def visitFugueSampleMethod(self, ctx: fp.FugueSampleMethodContext) -> Tuple:
        if ctx.rows is not None:
            n: Any = int(self.ctxToStr(ctx.rows))
        else:
            n = None
        if ctx.percentage is not None:
            frac: Any = float(self.ctxToStr(ctx.percentage)) / 100.0
        else:
            frac = None
        return n, frac

    def visitFugueZipType(self, ctx: fp.FugueZipTypeContext) -> str:
        return self.ctxToStr(ctx, delimit="_").lower()

    def visitFugueLoadColumns(self, ctx: fp.FugueLoadColumnsContext) -> Any:
        if ctx.schema is not None:
            return str(self.visit(ctx.schema))
        else:
            return self.visit(ctx.cols)

    def visitFugueSaveMode(self, ctx: fp.FugueSaveModeContext) -> str:
        mode = self.ctxToStr(ctx).lower()
        if mode == "to":
            mode = "error"
        return mode

    def visitFugueFileFormat(self, ctx: fp.FugueFileFormatContext) -> str:
        return self.ctxToStr(ctx).lower()

    def visitFuguePath(self, ctx: fp.FuguePathContext) -> Any:
        return eval(self.ctxToStr(ctx))

    def visitFugueCheckpointWeak(self, ctx: fp.FugueCheckpointWeakContext) -> Any:
        lazy = ctx.LAZY() is not None
        data = self.get_dict(ctx, "params")
        if "params" not in data:
            return lambda name, x: x.persist()
        else:
            return lambda name, x: x.weak_checkpoint(lazy=lazy, **data["params"])

    def visitFugueCheckpointStrong(self, ctx: fp.FugueCheckpointStrongContext) -> Any:
        lazy = ctx.LAZY() is not None
        data = self.get_dict(ctx, "partition", "single", "params")
        return lambda name, x: x.strong_checkpoint(
            lazy=lazy,
            partition=data.get("partition"),
            single="single" in data,
            **data.get("params", {}),
        )

    def visitFugueCheckpointDeterministic(
        self, ctx: fp.FugueCheckpointDeterministicContext
    ) -> Any:
        def _func(name: str, x: WorkflowDataFrame) -> WorkflowDataFrame:
            data = self.get_dict(ctx, "ns", "partition", "single", "params")

            x.deterministic_checkpoint(
                lazy=ctx.LAZY() is not None,
                partition=data.get("partition"),
                single="single" in data,
                namespace=data.get("ns"),
                **data.get("params", {}),
            )
            return x

        return _func

    def visitFugueYield(self, ctx: fp.FugueYieldContext) -> Any:
        def _func(name: str, x: WorkflowDataFrame) -> WorkflowDataFrame:
            yield_name = self.ctxToStr(ctx.name) if ctx.name is not None else name
            assert_or_throw(yield_name is not None, "yield name is not specified")
            if ctx.DATAFRAME() is None:
                if ctx.FILE() is not None:
                    x.yield_file_as(yield_name)
                elif ctx.TABLE() is not None:
                    x.yield_table_as(yield_name)
                else:  # pragma: no cover
                    raise NotImplementedError(self.ctxToStr(ctx))
            else:
                x.yield_dataframe_as(yield_name, as_local=ctx.LOCAL() is not None)
            return x

        return _func

    def visitFugueCheckpointNamespace(self, ctx: fp.FugueCheckpointNamespaceContext):
        return str(eval(self.ctxToStr(ctx)))


class _Extensions(_VisitorBase):
    def __init__(
        self,
        sql: FugueSQLParser,
        hooks: FugueSQLHooks,
        workflow: FugueWorkflow,
        dialect: str,
        variables: Optional[
            Dict[
                str, Tuple[WorkflowDataFrame, WorkflowDataFrames, LazyWorkflowDataFrame]
            ]
        ] = None,
        last: Optional[WorkflowDataFrame] = None,
        global_vars: Optional[Dict[str, Any]] = None,
        local_vars: Optional[Dict[str, Any]] = None,
    ):
        super().__init__(sql)
        self._workflow = workflow
        self._variables: Dict[
            str, Tuple[WorkflowDataFrame, WorkflowDataFrames, LazyWorkflowDataFrame]
        ] = {}
        if variables is not None:
            self._variables.update(variables)
        self._last: Optional[WorkflowDataFrame] = last
        self._hooks = hooks
        self._global_vars, self._local_vars = get_caller_global_local_vars(
            global_vars, local_vars
        )
        self._dialect = dialect

    @property
    def workflow(self) -> FugueWorkflow:
        return self._workflow

    @property
    def hooks(self) -> FugueSQLHooks:
        return self._hooks

    @property
    def variables(
        self,
    ) -> Dict[str, Tuple[WorkflowDataFrame, WorkflowDataFrames, LazyWorkflowDataFrame]]:
        return self._variables

    @property
    def last(self) -> WorkflowDataFrame:
        if self._last is None:
            raise FugueSQLError("latest dataframe does not exist")
        return self._last

    @property
    def global_vars(self) -> Dict[str, Any]:
        return self._global_vars

    @property
    def local_vars(self) -> Dict[str, Any]:
        return self._local_vars

    def get_df(
        self, key: str, ctx: fp.FugueDataFrameMemberContext
    ) -> WorkflowDataFrame:
        assert_or_throw(
            key in self.variables,
            lambda: FugueSQLSyntaxError(f"{key} is not defined"),
        )
        if isinstance(self.variables[key], LazyWorkflowDataFrame):
            assert_or_throw(
                ctx is None,
                FugueSQLSyntaxError("can't specify index or key for dataframe"),
            )
            return self.variables[key].get_df()  # type: ignore
        if isinstance(self.variables[key], WorkflowDataFrame):
            assert_or_throw(
                ctx is None,
                FugueSQLSyntaxError("can't specify index or key for dataframe"),
            )
            return self.variables[key]  # type: ignore
        assert_or_throw(
            ctx is not None,
            FugueSQLSyntaxError("must specify index or key for dataframes"),
        )
        if ctx.index is not None:
            return self.variables[key][int(self.ctxToStr(ctx.index))]  # type: ignore
        else:
            return self.variables[key][self.ctxToStr(ctx.key)]  # type: ignore

    def visitFugueDataFrameSource(
        self, ctx: fp.FugueDataFrameSourceContext
    ) -> WorkflowDataFrame:
        name = self.ctxToStr(ctx.fugueIdentifier(), delimit="")
        return self.get_df(name, ctx.fugueDataFrameMember())

    def visitFugueDataFrameNested(
        self, ctx: fp.FugueDataFrameNestedContext
    ) -> WorkflowDataFrame:
        sub = _Extensions(
            self.sql,
            self.hooks,
            workflow=self.workflow,
            dialect=self._dialect,
            variables=self.variables,
            last=self._last,
            global_vars=self.global_vars,
            local_vars=self.local_vars,
        )
        sub.visit(ctx.task)
        return sub.last

    def visitFugueDataFramePair(self, ctx: fp.FugueDataFramePairContext) -> Any:
        return self.to_kv(ctx)

    def visitFugueDataFramesList(
        self, ctx: fp.FugueDataFramesListContext
    ) -> WorkflowDataFrames:
        dfs = self.collectChildren(ctx, fp.FugueDataFrameContext)
        return WorkflowDataFrames(dfs)

    def visitFugueDataFramesDict(
        self, ctx: fp.FugueDataFramesDictContext
    ) -> WorkflowDataFrames:
        dfs = self.collectChildren(ctx, fp.FugueDataFramePairContext)
        return WorkflowDataFrames(dfs)

    def visitFugueTransformTask(
        self, ctx: fp.FugueTransformTaskContext
    ) -> WorkflowDataFrame:
        data = self.get_dict(ctx, "partition", "dfs", "params", "callback")
        if "dfs" not in data:
            data["dfs"] = WorkflowDataFrames(self.last)
        p = data["params"]
        using = _to_transformer(
            p["fugueUsing"],
            schema=p.get("schema"),
            global_vars=self.global_vars,
            local_vars=self.local_vars,
        )
        __modified_exception__ = self.to_runtime_error(ctx)  # noqa
        # TODO: ignore errors is not implemented
        return self.workflow.transform(
            data["dfs"],
            using=using,
            params=p.get("params"),
            pre_partition=data.get("partition"),
            callback=to_function(data["callback"], self.global_vars, self.local_vars)
            if "callback" in data
            else None,
        )

    def visitFugueOutputTransformTask(
        self, ctx: fp.FugueOutputTransformTaskContext
    ) -> None:
        data = self.get_dict(
            ctx, "partition", "dfs", "fugueUsing", "params", "callback"
        )
        if "dfs" not in data:
            data["dfs"] = WorkflowDataFrames(self.last)
        using = _to_output_transformer(
            data["fugueUsing"],
            global_vars=self.global_vars,
            local_vars=self.local_vars,
        )
        __modified_exception__ = self.to_runtime_error(ctx)  # noqa
        # TODO: ignore errors is not implemented
        self.workflow.out_transform(
            data["dfs"],
            using=using,
            params=data.get("params"),
            pre_partition=data.get("partition"),
            callback=to_function(data["callback"], self.global_vars, self.local_vars)
            if "callback" in data
            else None,
        )

    def visitFugueProcessTask(
        self, ctx: fp.FugueProcessTaskContext
    ) -> WorkflowDataFrame:
        data = self.get_dict(ctx, "partition", "dfs", "params")
        if "dfs" not in data:
            data["dfs"] = WorkflowDataFrames(self.last)
        p = data["params"]
        using = _to_processor(
            p["fugueUsing"],
            schema=p.get("schema"),
            global_vars=self.global_vars,
            local_vars=self.local_vars,
        )
        __modified_exception__ = self.to_runtime_error(ctx)  # noqa
        return self.workflow.process(
            data["dfs"],
            using=using,
            params=p.get("params"),
            pre_partition=data.get("partition"),
        )

    def visitFugueCreateTask(self, ctx: fp.FugueCreateTaskContext) -> WorkflowDataFrame:
        data = self.get_dict(ctx, "params")
        p = data["params"]
        using = _to_creator(
            p["fugueUsing"],
            schema=p.get("schema"),
            global_vars=self.global_vars,
            local_vars=self.local_vars,
        )
        __modified_exception__ = self.to_runtime_error(ctx)  # noqa
        return self.workflow.create(using=using, params=p.get("params"))

    def visitFugueCreateDataTask(
        self, ctx: fp.FugueCreateDataTaskContext
    ) -> WorkflowDataFrame:
        data = self.get_dict(ctx, "data", "schema")
        __modified_exception__ = self.to_runtime_error(ctx)  # noqa
        return self.workflow.df(
            data["data"], schema=data["schema"], data_determiner=to_uuid
        )

    def visitFugueZipTask(self, ctx: fp.FugueZipTaskContext) -> WorkflowDataFrame:
        data = self.get_dict(ctx, "dfs", "how")
        partition_spec = PartitionSpec(**self.get_dict(ctx, "by", "presort"))
        __modified_exception__ = self.to_runtime_error(ctx)  # noqa
        # TODO: currently SQL does not support cache to file on ZIP
        return self.workflow.zip(
            data["dfs"], how=data.get("how", "inner"), partition=partition_spec
        )

    def visitFugueOutputTask(self, ctx: fp.FugueOutputTaskContext):
        data = self.get_dict(ctx, "dfs", "fugueUsing", "params", "partition")
        if "dfs" not in data:
            data["dfs"] = WorkflowDataFrames(self.last)
        using = _to_outputter(
            data["fugueUsing"],
            global_vars=self.global_vars,
            local_vars=self.local_vars,
        )
        __modified_exception__ = self.to_runtime_error(ctx)  # noqa
        self.workflow.output(
            data["dfs"],
            using=using,
            params=data.get("params"),
            pre_partition=data.get("partition"),
        )

    def visitFuguePrintTask(self, ctx: fp.FuguePrintTaskContext) -> None:
        data = self.get_dict(ctx, "dfs")
        if "dfs" not in data:
            data["dfs"] = WorkflowDataFrames(self.last)
        params: Dict[str, Any] = {}
        if ctx.rows is not None:
            params["n"] = int(self.ctxToStr(ctx.rows))
        if ctx.count is not None:
            params["with_count"] = True
        if ctx.title is not None:
            params["title"] = eval(self.ctxToStr(ctx.title))
        __modified_exception__ = self.to_runtime_error(ctx)  # noqa
        self.workflow.show(data["dfs"], **params)

    def visitFugueSaveTask(self, ctx: fp.FugueSaveTaskContext):
        data = self.get_dict(
            ctx, "partition", "df", "m", "single", "fmt", "path", "params"
        )
        if "df" in data:
            df = data["df"]
        else:
            df = self.last
        df.save(
            path=data["path"],
            fmt=data.get("fmt", ""),
            mode=data["m"],
            partition=data.get("partition"),
            single="single" in data,
            **data.get("params", {}),
        )

    def visitFugueSaveAndUseTask(self, ctx: fp.FugueSaveAndUseTaskContext):
        data = self.get_dict(
            ctx, "partition", "df", "m", "single", "fmt", "path", "params"
        )
        if "df" in data:
            df = data["df"]
        else:
            df = self.last
        return df.save_and_use(
            path=data["path"],
            fmt=data.get("fmt", ""),
            mode=data["m"],
            partition=data.get("partition"),
            single="single" in data,
            **data.get("params", {}),
        )

    def visitFugueRenameColumnsTask(self, ctx: fp.FugueRenameColumnsTaskContext):
        data = self.get_dict(ctx, "cols", "df")
        if "df" in data:
            df = data["df"]
        else:
            df = self.last
        return df.rename(data["cols"])

    def visitFugueAlterColumnsTask(self, ctx: fp.FugueAlterColumnsTaskContext):
        data = self.get_dict(ctx, "cols", "df")
        if "df" in data:
            df = data["df"]
        else:
            df = self.last
        return df.alter_columns(data["cols"])

    def visitFugueDropColumnsTask(self, ctx: fp.FugueDropColumnsTaskContext):
        data = self.get_dict(ctx, "cols", "df")
        if "df" in data:
            df = data["df"]
        else:
            df = self.last
        return df.drop(
            columns=data["cols"],
            if_exists=ctx.IF() is not None,
        )

    def visitFugueDropnaTask(self, ctx: fp.FugueDropnaTaskContext):
        data = self.get_dict(ctx, "cols", "df")
        if "df" in data:
            df = data["df"]
        else:
            df = self.last
        params: Dict[str, Any] = {}
        params["how"] = "any" if ctx.ANY() is not None else "all"
        if "cols" in data:
            params["subset"] = data["cols"]
        return df.dropna(**params)

    def visitFugueFillnaTask(self, ctx: fp.FugueFillnaTaskContext):
        data = self.get_dict(ctx, "params", "df")
        if "df" in data:
            df = data["df"]
        else:
            df = self.last
        return df.fillna(value=data["params"])

    def visitFugueSampleTask(self, ctx: fp.FugueSampleTaskContext):
        data = self.get_dict(ctx, "df")
        if "df" in data:
            df = data["df"]
        else:
            df = self.last
        params: Dict[str, Any] = {}
        params["replace"] = ctx.REPLACE() is not None
        if ctx.seed is not None:
            params["seed"] = int(self.ctxToStr(ctx.seed))
        n, frac = self.visit(ctx.method)
        if n is not None:
            params["n"] = n
        if frac is not None:
            params["frac"] = frac
        return df.sample(**params)

    def visitFugueTakeTask(self, ctx: fp.FugueTakeTaskContext):
        data = self.get_dict(ctx, "partition", "presort", "df")
        if "df" in data:
            df = data["df"]
        else:
            df = self.last
        params: Dict[str, Any] = {}
        params["n"] = int(self.ctxToStr(ctx.rows)) or 20  # default is 20
        params["na_position"] = "first" if ctx.FIRST() is not None else "last"
        if data.get("partition"):
            _partition_spec = PartitionSpec(data.get("partition"))
            return df.partition(
                by=_partition_spec.partition_by, presort=_partition_spec.presort
            ).take(**params)
        else:
            if data.get("presort"):
                params["presort"] = data.get("presort")
            return df.take(**params)

    def visitFugueLoadTask(self, ctx: fp.FugueLoadTaskContext) -> WorkflowDataFrame:
        data = self.get_dict(ctx, "fmt", "path", "params", "columns")
        __modified_exception__ = self.to_runtime_error(ctx)  # noqa
        return self.workflow.load(
            path=data["path"],
            fmt=data.get("fmt", ""),
            columns=data.get("columns"),
            **data.get("params", {}),
        )

    def visitFugueNestableTask(self, ctx: fp.FugueNestableTaskContext) -> None:
        data = self.get_dict(ctx, "q")
        statements = list(self._beautify_sql(data["q"]))
        if len(statements) == 1 and isinstance(statements[0], WorkflowDataFrame):
            df: Any = statements[0]
        else:
            __modified_exception__ = self.to_runtime_error(ctx)  # noqa
            df = self.workflow.select(*statements, dialect=self._dialect)
        self._process_assignable(df, ctx)

    def visitFugueModuleTask(self, ctx: fp.FugueModuleTaskContext) -> None:
        data = self.get_dict(ctx, "assign", "dfs", "fugueUsing", "params")
        sub = _to_module(
            data["fugueUsing"],
            global_vars=self.global_vars,
            local_vars=self.local_vars,
        )
        varname = data["assign"][0] if "assign" in data else None
        if varname is not None:
            assert_or_throw(
                sub.has_single_output or sub.has_multiple_output,
                FugueSQLSyntaxError("invalid assignment for module without output"),
            )
        if sub.has_input:
            dfs = data["dfs"] if "dfs" in data else WorkflowDataFrames(self.last)
        else:
            dfs = WorkflowDataFrames()
        p = data["params"] if "params" in data else {}
        if sub.has_dfs_input:
            result = sub(dfs, **p)
        elif len(dfs) == 0:
            result = sub(self.workflow, **p)
        elif len(dfs) == 1 or not dfs.has_key:
            result = sub(*list(dfs.values()), **p)
        else:
            result = sub(**dfs, **p)
        if sub.has_single_output or sub.has_multiple_output:
            self.variables[varname] = result
        if sub.has_single_output:
            self._last = result

    def visitFugueSqlEngine(
        self, ctx: fp.FugueSqlEngineContext
    ) -> Tuple[Any, Dict[str, Any]]:
        data = self.get_dict(ctx, "fugueUsing", "params")
        try:
            engine: Any = to_type(
                data["fugueUsing"],
                SQLEngine,
                global_vars=self.global_vars,
                local_vars=self.local_vars,
            )
        except TypeError:
            engine = str(data["fugueUsing"])
        return engine, data.get("params", {})

    def visitQuery(self, ctx: fp.QueryContext) -> Iterable[Any]:
        def get_sql() -> str:
            return " ".join(
                [
                    "" if ctx.ctes() is None else self.ctxToStr(ctx.ctes()),
                    self.ctxToStr(ctx.queryTerm()),
                    self.ctxToStr(ctx.queryOrganization()),
                ]
            ).strip()

        if ctx.fugueSqlEngine() is not None:
            engine, engine_params = self.visitFugueSqlEngine(ctx.fugueSqlEngine())
            __modified_exception__ = self.to_runtime_error(ctx)  # noqa
            yield self.workflow.select(
                get_sql(),
                sql_engine=engine,
                sql_engine_params=engine_params,
                dialect=self._dialect,
            )
        elif ctx.ctes() is None:
            yield from self._get_query_elements(ctx)
        else:
            __modified_exception__ = self.to_runtime_error(ctx)  # noqa
            yield self.workflow.select(get_sql(), dialect=self._dialect)

    def visitOptionalFromClause(
        self, ctx: fp.OptionalFromClauseContext
    ) -> Iterable[Any]:
        c = ctx.fromClause()
        if c is None:
            yield "FROM"
            yield self.last
        else:
            yield from self._get_query_elements(ctx)

    def visitTableName(self, ctx: fp.TableNameContext) -> Iterable[Any]:
        table_name = self.ctxToStr(ctx.multipartIdentifier(), delimit="")
        if table_name not in self.variables:
            assert_or_throw(
                ctx.fugueDataFrameMember() is None,
                FugueSQLSyntaxError("can't specify index or key for dataframe"),
            )
            table: Any = self.hooks.on_select_source_not_found(
                self.workflow, table_name
            )
        else:
            table = self.get_df(table_name, ctx.fugueDataFrameMember())
        if isinstance(table, str):
            yield table
            yield from self._get_query_elements(ctx.sample())
            yield from self._get_query_elements(ctx.tableAlias())
        else:
            yield table
            yield from self._get_query_elements(ctx.sample())
            if ctx.tableAlias().strictIdentifier() is not None:
                yield from self._get_query_elements(ctx.tableAlias())
            elif validate_triad_var_name(table_name):
                yield "AS"
                yield table_name

    def visitFugueNestableTaskCollectionNoSelect(
        self, ctx: fp.FugueNestableTaskCollectionNoSelectContext
    ) -> Iterable[Any]:
        last = self._last
        for i in range(ctx.getChildCount()):
            n = ctx.getChild(i)
            sub = _Extensions(
                self.sql,
                self.hooks,
                workflow=self.workflow,
                dialect=self._dialect,
                variables=self.variables,
                last=last,
                global_vars=self.global_vars,
                local_vars=self.local_vars,
            )
            yield sub.visit(n)

    def visitSetOperation(self, ctx: fp.SetOperationContext) -> Iterable[Any]:
        def get_sub(_ctx: Tree) -> List[Any]:

            sub = list(
                self.visitFugueTerm(_ctx)
                if isinstance(_ctx, fp.FugueTermContext)
                else self._get_query_elements(_ctx)
            )
            if len(sub) == 1 and isinstance(sub[0], WorkflowDataFrame):
                return ["SELECT * FROM", sub[0]]
            else:
                return sub

        yield from get_sub(ctx.left)
        yield from self._get_query_elements(ctx.theOperator)
        if ctx.setQuantifier() is not None:
            yield from self._get_query_elements(ctx.setQuantifier())
        yield from get_sub(ctx.right)

    def visitAliasedQuery(self, ctx: fp.AliasedQueryContext) -> Iterable[Any]:
        sub = list(self._get_query_elements(ctx.query()))
        if len(sub) == 1 and isinstance(sub[0], WorkflowDataFrame):
            yield sub[0]
        else:
            yield "("
            yield from sub
            yield ")"
        yield from self._get_query_elements(ctx.sample())
        yield from self._get_query_elements(ctx.tableAlias())

    def _beautify_sql(self, statements: Iterable[Any]) -> Iterable[Any]:
        current = ""
        for s in statements:
            if not isinstance(s, str):
                if current != "":
                    yield current
                yield s
                current = ""
            else:
                s = s.strip()
                if s != "":
                    if (
                        current == ""
                        or current.endswith(".")
                        or s.startswith(".")
                        or current.endswith("(")
                        or s.startswith(")")
                    ):
                        current += s
                    else:
                        current += " " + s
        if current != "":
            yield current

    def _get_query_elements(self, node: Tree) -> Iterable[Any]:  # noqa: C901
        if node is None:
            return
        if isinstance(node, CommonToken):
            yield self.sql.code[node.start : node.stop + 1]
            return
        if isinstance(node, TerminalNode):
            token = node.getSymbol()
            yield self.sql.code[token.start : token.stop + 1]
        for i in range(node.getChildCount()):
            n = node.getChild(i)
            if isinstance(n, fp.TableNameContext):
                yield from self.visitTableName(n)
            elif isinstance(n, fp.OptionalFromClauseContext):
                yield from self.visitOptionalFromClause(n)
            elif isinstance(n, fp.FugueTermContext):
                yield from self.visitFugueTerm(n)
            elif isinstance(n, fp.AliasedQueryContext):
                yield from self.visitAliasedQuery(n)
            elif isinstance(n, fp.SetOperationContext):
                yield from self.visitSetOperation(n)
            else:
                yield from self._get_query_elements(n)

    def _process_assignable(self, df: WorkflowDataFrame, ctx: Tree):
        data = self.get_dict(ctx, "assign", "checkpoint", "broadcast", "y")
        if "assign" in data:
            varname, _ = data["assign"]
        else:
            varname = None
        if "checkpoint" in data:
            data["checkpoint"](varname, df)
        if "broadcast" in data:
            df = df.broadcast()
        if "y" in data:
            data["y"](varname, df)
        if varname is not None:
            self.variables[varname] = df  # type: ignore
        self._last = df



import copy
import os
from random import randint
from typing import Any, Dict, Iterable, List, Optional

import pandas as pd
from adagio.instances import WorkflowResultCache
from pytest import raises
from triad.exceptions import InvalidOperationError

from fugue import (
    ArrayDataFrame,
    DataFrame,
    FugueWorkflow,
    NativeExecutionEngine,
    QPDPandasEngine,
    PandasDataFrame,
    WorkflowDataFrames,
)
from fugue.collections.partition import PartitionSpec
from fugue.collections.yielded import PhysicalYielded
from fugue.constants import (
    FUGUE_CONF_WORKFLOW_CHECKPOINT_PATH,
    FUGUE_CONF_WORKFLOW_EXCEPTION_HIDE,
)
from fugue.dataframe.utils import _df_eq as df_eq
from fugue.exceptions import FuguePluginsRegistrationError, FugueWorkflowCompileError
from fugue.execution import NativeExecutionEngine
from fugue.extensions.transformer.convert import transformer


def test_worflow_dataframes():
    dag1 = FugueWorkflow()
    df1 = dag1.df([[0]], "a:int")
    df2 = dag1.df([[0]], "b:int")
    dag2 = FugueWorkflow()
    df3 = dag2.df([[0]], "a:int")

    dfs1 = WorkflowDataFrames(a=df1, b=df2)
    assert dfs1["a"] is df1
    assert dfs1["b"] is df2

    dfs2 = WorkflowDataFrames(dfs1, aa=df1, bb=df2)
    assert 4 == len(dfs2)

    with raises(ValueError):
        WorkflowDataFrames(a=df1, b=df3)

    with raises(ValueError):
        WorkflowDataFrames(a=df1, b=ArrayDataFrame([[0]], "a:int"))

    dag = FugueWorkflow()
    df = dag.df([[0, 1], [1, 1]], "a:int,b:int")
    assert df.partition_spec.empty
    df2 = df.partition(by=["a"])
    assert df.partition_spec.empty
    assert df2.partition_spec == PartitionSpec(by=["a"])
    df3 = df.partition_by("a", "b")
    assert df.partition_spec.empty
    assert df3.partition_spec == PartitionSpec(by=["a", "b"])
    df4 = df.per_partition_by("a", "b")
    assert df.partition_spec.empty
    assert df4.partition_spec == PartitionSpec(by=["a", "b"], algo="even")
    df4 = df.per_row()
    assert df.partition_spec.empty
    assert df4.partition_spec == PartitionSpec("per_row")


def test_workflow():
    builder = FugueWorkflow()

    a = builder.create_data([[0], [0], [1]], "a:int")
    assert builder.last_df is a
    raises(InvalidOperationError, lambda: a._task.copy())
    raises(InvalidOperationError, lambda: copy.copy(a._task))
    raises(InvalidOperationError, lambda: copy.deepcopy(a._task))
    a.show()
    a.show()

    raises(FugueWorkflowCompileError, lambda: builder.df(123))

    b = a.transform(mock_tf1, "*,b:int", pre_partition=dict(by=["a"]))
    b.show()
    assert builder.last_df is b
    builder.create_data([[0], [1]], "b:int").show()
    assert builder.last_df is not b
    c = ArrayDataFrame([[100]], "a:int")
    builder.show(a, b, c)
    b = a.partition(by=["a"]).transform(mock_tf2).persist().broadcast()
    b.show()
    assert builder.last_df is b
    c = builder.df(a)
    assert builder.last_df is a

    builder.run()
    df_eq(a.result, [[0], [0], [1]], "a:int")
    raises(FuguePluginsRegistrationError, lambda: builder.run("abc"))

    builder.run("NativeExecutionEngine")
    df_eq(b.result, [[0, 2], [0, 2], [1, 1]], "a:int,b:int")
    df_eq(b.compute(), [[0, 2], [0, 2], [1, 1]], "a:int,b:int")
    df_eq(b.compute(NativeExecutionEngine), [[0, 2], [0, 2], [1, 1]], "a:int,b:int")


def test_yield(tmpdir):
    df = pd.DataFrame([[0, 0]], columns=["a", "b"])

    # schema: *
    def t(df: pd.DataFrame) -> pd.DataFrame:
        return df.assign(b=df.b + 1)

    dag = FugueWorkflow()
    dag.df(df).transform(t).yield_dataframe_as("x")
    result = dag.run()["x"]
    assert [[0, 1]] == result.as_array()

    dag1 = FugueWorkflow()
    dag1.df(df).transform(t).yield_file_as("x")
    res = dag1.run("", {FUGUE_CONF_WORKFLOW_CHECKPOINT_PATH: str(tmpdir)})
    assert isinstance(res.yields["x"], PhysicalYielded)
    assert res.yields["x"].storage_type == "file"

    dag2 = FugueWorkflow()
    dag2.df(dag1.yields["x"]).transform(t).yield_dataframe_as("y")
    result = dag2.run("", {FUGUE_CONF_WORKFLOW_CHECKPOINT_PATH: str(tmpdir)})["y"]
    assert [[0, 2]] == result.as_array()

    dag3 = FugueWorkflow()
    dag3.df(dag2.yields["y"]).transform(t).yield_dataframe_as("z")
    result = dag3.run()["z"]
    assert [[0, 3]] == result.as_array()

    dag4 = FugueWorkflow()
    dag4.df(df).transform(t).yield_table_as("x")
    res = dag4.run(MockEngine, {FUGUE_CONF_WORKFLOW_CHECKPOINT_PATH: str(tmpdir)})
    assert isinstance(res.yields["x"], PhysicalYielded)
    assert res.yields["x"].storage_type == "table"

    dag5 = FugueWorkflow()
    dag5.df(dag4.yields["x"]).transform(t).yield_dataframe_as("y")
    result = dag5.run(MockEngine, {FUGUE_CONF_WORKFLOW_CHECKPOINT_PATH: str(tmpdir)})[
        "y"
    ]
    assert [[0, 2]] == result.as_array()


def test_compile_conf():
    dag = FugueWorkflow(compile_conf={"a": 1})
    assert dag.conf["a"] == 1

    dag = FugueWorkflow()  # default conf test
    assert dag.conf.get_or_throw(FUGUE_CONF_WORKFLOW_EXCEPTION_HIDE, str) != ""


class MockSQLEngine(QPDPandasEngine):
    def table_exists(self, table: str) -> bool:
        path = os.path.join(
            self.conf[FUGUE_CONF_WORKFLOW_CHECKPOINT_PATH], table + ".parquet"
        )
        return os.path.exists(path)

    def save_table(
        self,
        df: DataFrame,
        table: str,
        mode: str = "overwrite",
        partition_spec: Optional[PartitionSpec] = None,
        **kwargs: Any
    ) -> None:
        path = os.path.join(
            self.conf[FUGUE_CONF_WORKFLOW_CHECKPOINT_PATH], table + ".parquet"
        )
        df.as_pandas().to_parquet(path)

    def load_table(self, table: str, **kwargs: Any) -> DataFrame:
        path = os.path.join(
            self.conf[FUGUE_CONF_WORKFLOW_CHECKPOINT_PATH], table + ".parquet"
        )
        return PandasDataFrame(pd.read_parquet(path))


class MockEngine(NativeExecutionEngine):
    def create_default_sql_engine(self):
        return MockSQLEngine(self)


class MockCache(WorkflowResultCache):
    def __init__(self, ctx=None, dummy=True):
        self.dummy = dummy
        self.tb = dict()
        self.set_called = 0
        self.skip_called = 0
        self.get_called = 0
        self.hit = 0

    def set(self, key: str, value: Any) -> None:
        self.tb[key] = (False, value)
        print("set", key)
        self.set_called += 1

    def skip(self, key: str) -> None:
        self.tb[key] = (True, None)
        self.skip_called += 1

    def get(self, key: str):
        if self.dummy:
            return True, False, ArrayDataFrame([[100]], "a:int")
        self.get_called += 1
        if key not in self.tb:
            print("not get", key)
            return False, False, None
        x = self.tb[key]
        print("get", key)
        self.hit += 1
        return True, x[0], x[1]


def mock_tf1(df: List[Dict[str, Any]], v: int = 1) -> Iterable[Dict[str, Any]]:
    for r in df:
        r["b"] = v * len(df)
        yield r


@transformer("*,b:int")
def mock_tf2(df: List[Dict[str, Any]], v: int = 1) -> Iterable[Dict[str, Any]]:
    for r in df:
        r["b"] = v * len(df)
        yield r


# schema: a:int
def create_rand() -> List[List[Any]]:
    return [[randint(0, 10)]]


def my_show(df: DataFrame) -> DataFrame:
    df.show()
    return df
